# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import sys
import warnings
from pathlib import Path
from typing import Optional, Union


def get_logger(
    name: str,
    level: str = "info",
    log_dir: Optional[Union[str, Path]] = None,
) -> logging.Logger:
    """Gets a logger instance with the specified name and log level.

    Args:
        name : The name of the logger.
        level (optional): The log level to set for the logger. Defaults to "info".
        log_dir (optional): Directory to store log files. Defaults to None.

    Returns:
        logging.Logger: The logger instance.
    """
    if name not in logging.Logger.manager.loggerDict:
        configure_logger(name, level=level, log_dir=log_dir)

    logger = logging.getLogger(name)
    return logger


def _detect_rank() -> int:
    """Detects rank.

    Reading the rank from the environment variables instead of
    get_device_rank_info to avoid circular imports.
    """
    for var_name in (
        "RANK",
        "SKYPILOT_NODE_RANK",  # SkyPilot
        "PMI_RANK",  # HPC
    ):
        rank = os.environ.get(var_name, None)
        if rank is not None:
            rank = int(rank)
            if rank < 0:
                raise ValueError(f"Negative rank: {rank} specified in '{var_name}'!")
            return rank
    return 0


def configure_logger(
    name: str,
    level: str = "info",
    log_dir: Optional[Union[str, Path]] = None,
) -> None:
    """Configures a logger with the specified name and log level."""
    logger = logging.getLogger(name)

    # Remove any existing handlers
    logger.handlers = []

    # Configure the logger
    logger.setLevel(level.upper())

    device_rank = _detect_rank()

    default_formatter = logging.Formatter(
        "[%(asctime)s][%(name)s]"
        f"[rank{device_rank}]"
        "[pid:%(process)d][%(threadName)s]"
        "[%(levelname)s]][%(filename)s:%(lineno)s] %(message)s"
    )

    # Add a console handler to the logger for only global leader.
    if device_rank == 0:
        if should_use_rich_logging():
            console_handler = _configure_rich_handler(device_rank, level)
        else:
            console_handler = logging.StreamHandler(sys.stdout)
            console_handler.setFormatter(default_formatter)

        console_handler.setLevel(level.upper())
        logger.addHandler(console_handler)

    # Add a file handler if log_dir is provided
    if log_dir:
        log_dir = Path(log_dir)
        log_dir.mkdir(parents=True, exist_ok=True)

        file_handler = logging.FileHandler(log_dir / f"rank_{device_rank:04d}.log")
        file_handler.setFormatter(default_formatter)
        file_handler.setLevel(level.upper())
        logger.addHandler(file_handler)

    logger.propagate = False


def should_use_rich_logging() -> bool:
    """Determines whether rich logging should be used.

    Returns:
        bool: True if rich logging should be used, False otherwise.

    Rich logging is enabled if the output is a terminal (TTY) and not explicitly
    disabled via the OUMI_DISABLE_RICH_LOGGING environment variable.
    """
    # Check if explicitly disabled
    if os.environ.get("OUMI_DISABLE_RICH_LOGGING", "").lower() in (
        "1",
        "yes",
        "on",
        "true",
        "y",
    ):
        return False

    return sys.stdout.isatty()  # is in a terminal


def _configure_rich_handler(
    device_rank: int,
    level: str,
) -> logging.Handler:
    """Configures a rich logging handler."""
    try:
        from rich.console import Console
        from rich.logging import RichHandler
        from rich.traceback import install
    except ImportError:
        raise ImportError(
            "Rich logging is not installed. Please install it with `pip install rich`."
        )

    use_detailed_logging = level.upper() == "DEBUG"

    if use_detailed_logging:
        # Add extra logging for debugging
        install(show_locals=True, suppress=[])

    console = Console()
    console_handler = RichHandler(
        console=console,
        show_time=True,
        show_level=True,
        show_path=True,
        enable_link_path=True,
        markup=False,
        rich_tracebacks=use_detailed_logging,
        tracebacks_show_locals=use_detailed_logging,
        locals_max_length=20,
        locals_max_string=80,
    )

    if use_detailed_logging:
        rich_formatter = logging.Formatter(
            f"[rank-{device_rank}][pid-%(process)d][%(threadName)s] %(message)s"
        )
    else:
        rich_formatter = logging.Formatter(f"[rank-{device_rank}] %(message)s")
    console_handler.setFormatter(rich_formatter)
    return console_handler


def update_logger_level(name: str, level: str = "info") -> None:
    """Updates the log level of the logger.

    Args:
        name (str): The logger instance to update.
        level (str, optional): The log level to set for the logger. Defaults to "info".
    """
    logger = get_logger(name, level=level)
    logger.setLevel(level.upper())

    for handler in logger.handlers:
        handler.setLevel(level.upper())


def configure_dependency_warnings(level: Union[str, int] = "info") -> None:
    """Ignores non-critical warnings from dependencies, unless in debug mode.

    Args:
        level (str, optional): The log level to set for the logger. Defaults to "info".
    """
    level_value = logging.DEBUG
    if isinstance(level, str):
        level_value = logging.getLevelName(level.upper())
        if not isinstance(level_value, int):
            raise TypeError(
                f"getLevelName() mapped log level name to non-integer: "
                f"{type(level_value)}!"
            )
    elif isinstance(level, int):
        level_value = int(level)

    if level_value > logging.DEBUG:
        warnings.filterwarnings(action="ignore", category=UserWarning, module="torch")
        warnings.filterwarnings(
            action="ignore", category=UserWarning, module="huggingface_hub"
        )
        warnings.filterwarnings(
            action="ignore", category=UserWarning, module="transformers"
        )


# Default logger for the package
logger = get_logger("oumi")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Annotated

import typer

import oumi.cli.cli_utils as cli_utils
from oumi.cli.alias import AliasType, try_get_config_name_for_alias
from oumi.utils.logging import logger


def tune(
    ctx: typer.Context,
    config: Annotated[
        str,
        typer.Option(
            *cli_utils.CONFIG_FLAGS, help="Path to the configuration file for tuning."
        ),
    ],
    level: cli_utils.LOG_LEVEL_TYPE = None,
    verbose: cli_utils.VERBOSE_TYPE = False,
):
    """Tune hyperparameters for a model.

    Args:
        ctx: The Typer context object.
        config: Path to the configuration file for training.
        level: The logging level for the specified command.
        verbose: Enable verbose logging with additional debug information.
    """
    extra_args = cli_utils.parse_extra_cli_args(ctx)

    config = str(
        cli_utils.resolve_and_fetch_config(
            try_get_config_name_for_alias(config, AliasType.TUNE),
        )
    )
    with cli_utils.CONSOLE.status(
        "[green]Loading configuration...[/green]", spinner="dots"
    ):
        # Delayed imports
        from oumi.core.configs import TuningConfig
        from oumi.tune import tune as oumi_tune
        from oumi.utils.torch_utils import (
            device_cleanup,
            limit_per_process_memory,
        )
        # End imports

    cli_utils.configure_common_env_vars()

    parsed_config: TuningConfig = TuningConfig.from_yaml_and_arg_list(
        config, extra_args, logger=logger
    )
    parsed_config.finalize_and_validate()

    limit_per_process_memory()
    device_cleanup()

    # Run training
    oumi_tune(parsed_config, verbose=verbose)

    device_cleanup()



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Optional

from oumi.core.configs.params.base_params import BaseParams


class EvaluationBackend(Enum):
    """Enum representing the evaluation backend to use."""

    LM_HARNESS = "lm_harness"
    ALPACA_EVAL = "alpaca_eval"
    CUSTOM = "custom"


@dataclass
class EvaluationTaskParams(BaseParams):
    """Configuration parameters for model evaluation tasks.

    Supported backends:

    - LM Harness: Framework for evaluating language models on standard benchmarks.
      A list of all supported tasks can be found at:
      https://github.com/EleutherAI/lm-evaluation-harness/tree/main/lm_eval/tasks.
    - Alpaca Eval: Framework for evaluating language models on instruction-following
      and quality of responses on open-ended questions.
    - Custom: Users can register their own evaluation functions using the decorator
      `@register_evaluation_function`. The task_name should be the registry key for
      the custom evaluation function to be used.

    Examples:
        .. code-block:: python

            # LM Harness evaluation on MMLU
            params = EvaluationTaskParams(
                evaluation_backend="lm_harness",
                task_name="mmlu",
                eval_kwargs={"num_fewshot": 5}
            )


        .. code-block:: python

            # Alpaca Eval 2.0 evaluation
            params = EvaluationTaskParams(
                evaluation_backend="alpaca_eval"
            )


        .. code-block:: python

            # Custom evaluation
            @register_evaluation_function("my_evaluation_function")
            def my_evaluation(task_params, config):
                accuracy = ...
                return EvaluationResult(task_result={"accuracy": accuracy})

            params = EvaluationTaskParams(
                task_name="my_evaluation_function",
                evaluation_backend="custom"
            )
    """

    evaluation_backend: str = ""
    """The evaluation backend to use for the current task."""

    task_name: Optional[str] = None
    """The task to evaluate or the custom evaluation function to use.

    For LM Harness evaluations (when the evaluation_backend is set to
    EvaluationBackend.LM_HARNESS), the `task_name` corresponds to a predefined task
    to evaluate on (e.g. "mmlu"). A list of all supported tasks by the LM Harness
    backend can be found by running: `lm-eval --tasks list`.

    For custom evaluations (when evaluation_backend is set to EvaluationBackend.CUSTOM),
    the `task_name` should be the registry key for the custom evaluation function to be
    used. Users can register new evaluation functions using the decorator
    `@register_evaluation_function`.
    """

    num_samples: Optional[int] = None
    """Number of samples/examples to evaluate from this dataset.

    Mostly for debugging, in order to reduce the runtime.
    If not set (None): the entire dataset is evaluated.
    If set, this must be a positive integer.
    """

    log_samples: Optional[bool] = False
    """Whether to log the samples used for evaluation.

    If not set (False): the model samples used for evaluation will not be logged.
    If set to True: the model samples generated during inference and used for
    evaluation will be logged in `backend_config.json`. The backend may also log
    other intermediate results related to inference.
    """

    eval_kwargs: dict[str, Any] = field(default_factory=dict)
    """Additional keyword arguments to pass to the evaluation function.

    This allows for passing any evaluation-specific parameters that are not
    covered by other fields in TaskParams classes.
    """

    def get_evaluation_backend(self) -> EvaluationBackend:
        """Returns the evaluation backend as an Enum."""
        if not self.evaluation_backend:
            raise ValueError(
                "Missing `evaluation_backend`. When running evaluations, it is "
                "necessary to specify the evaluation backend to use for EACH task. "
                "The available backends can be found in the following enum: "
                "`oumi.core.configs.params.evaluation_params.EvaluationBackend`. "
                f"Current options: {EvaluationTaskParams.list_evaluation_backends()}."
            )
        elif self.evaluation_backend == EvaluationBackend.LM_HARNESS.value:
            return EvaluationBackend.LM_HARNESS
        elif self.evaluation_backend == EvaluationBackend.ALPACA_EVAL.value:
            return EvaluationBackend.ALPACA_EVAL
        elif self.evaluation_backend == EvaluationBackend.CUSTOM.value:
            return EvaluationBackend.CUSTOM
        else:
            raise ValueError(f"Unknown evaluation backend: {self.evaluation_backend}")

    @staticmethod
    def list_evaluation_backends() -> str:
        """Returns a string listing all available evaluation backends."""
        return ", ".join([backend.value for backend in EvaluationBackend])

    def __post_init__(self):
        """Verifies params."""
        if self.num_samples is not None and self.num_samples <= 0:
            raise ValueError("`num_samples` must be None or a positive integer.")


@dataclass
class LMHarnessTaskParams(EvaluationTaskParams):
    """Parameters for the LM Harness evaluation framework.

    LM Harness is a comprehensive benchmarking suite for evaluating language models
    across various tasks.
    """

    num_fewshot: Optional[int] = None
    """Number of few-shot examples (with responses) to add in the prompt, in order to
    teach the model how to respond to the specific dataset's prompts.

    If not set (None): LM Harness will decide the value.
    If set to 0: no few-shot examples will be added in the prompt.
    """

    def __post_init__(self):
        """Verifies params."""
        if not self.task_name:
            raise ValueError("`task_name` must be a valid LM Harness task.")
        if self.num_fewshot and self.num_fewshot < 0:
            raise ValueError("`num_fewshot` must be non-negative.")


@dataclass
class AlpacaEvalTaskParams(EvaluationTaskParams):
    """Parameters for the AlpacaEval evaluation framework.

    AlpacaEval is an LLM-based automatic evaluation suite that is fast, cheap,
    replicable, and validated against 20K human annotations. The latest version
    (AlpacaEval 2.0) contains 805 prompts (tatsu-lab/alpaca_eval), which are open-ended
    questions. A model annotator (judge) is used to evaluate the quality of model's
    responses for these questions and calculates win rates vs. reference responses.
    The default judge is GPT4 Turbo.
    """

    version: Optional[float] = 2.0
    """The version of AlpacaEval to use. Options: 1.0 or 2.0 (default)."""

    def __post_init__(self):
        """Verifies params."""
        if self.version not in [1.0, 2.0]:
            raise ValueError("AlpacaEval `version` must be 1.0 or 2.0.")



import tempfile
from pathlib import Path
from unittest.mock import ANY, MagicMock, Mock, patch

import jsonlines
import PIL.Image
import pytest

from oumi.core.configs import GenerationParams, InferenceConfig, ModelParams
from oumi.core.configs.params.guided_decoding_params import GuidedDecodingParams
from oumi.core.types.conversation import ContentItem, Conversation, Message, Role, Type
from oumi.inference import VLLMInferenceEngine
from oumi.utils.conversation_utils import base64encode_content_item_image_bytes
from oumi.utils.image_utils import (
    create_png_bytes_from_image,
)

try:
    vllm_import_failed = False
    from vllm.lora.request import LoRARequest  # type: ignore
    from vllm.outputs import (  # pyright: ignore[reportMissingImports]
        CompletionOutput,
        RequestOutput,
    )

    def _create_vllm_output(responses: list[str], output_id: str) -> RequestOutput:
        outputs = []
        for ind, response in enumerate(responses):
            outputs.append(
                CompletionOutput(
                    text=response,
                    index=ind,
                    token_ids=[],
                    cumulative_logprob=None,
                    logprobs=None,
                )
            )
        return RequestOutput(
            request_id=output_id,
            outputs=outputs,
            prompt=None,
            prompt_token_ids=[],
            prompt_logprobs=None,
            finished=True,
        )
except ModuleNotFoundError:
    vllm_import_failed = True


@pytest.fixture
def mock_sampling_params():
    with patch("oumi.inference.vllm_inference_engine.SamplingParams") as mock:
        yield mock


#
# Fixtures
#
@pytest.fixture
def mock_vllm():
    with patch("oumi.inference.vllm_inference_engine.vllm") as mvllm:
        yield mvllm


@pytest.fixture
def mock_lora_request():
    with patch("oumi.inference.vllm_inference_engine.LoRARequest") as mlo:
        yield mlo


def _get_default_model_params(use_lora: bool = False) -> ModelParams:
    return ModelParams(
        model_name="MlpEncoder",
        adapter_model="/path/to/adapter" if use_lora else None,
        trust_remote_code=True,
        tokenizer_pad_token="<pad>",
        tokenizer_name="gpt2",
    )


def _get_default_inference_config() -> InferenceConfig:
    return InferenceConfig(generation=GenerationParams(max_new_tokens=5))


def _setup_input_conversations(filepath: str, conversations: list[Conversation]):
    Path(filepath).parent.mkdir(parents=True, exist_ok=True)
    Path(filepath).touch()
    with jsonlines.open(filepath, mode="w") as writer:
        for conversation in conversations:
            json_obj = conversation.to_dict()
            writer.write(json_obj)
    # Add some empty lines into the file
    with open(filepath, "a") as f:
        f.write("\n\n\n")


def _create_test_pil_image() -> PIL.Image.Image:
    pil_image = PIL.Image.new(mode="RGB", size=(32, 48))
    return pil_image


def _create_test_png_image_bytes() -> bytes:
    return create_png_bytes_from_image(_create_test_pil_image())


def _create_test_png_image_base64_str() -> str:
    return base64encode_content_item_image_bytes(
        ContentItem(binary=_create_test_png_image_bytes(), type=Type.IMAGE_BINARY),
        add_mime_prefix=True,
    )


#
# Tests
#
@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_infer_online(mock_vllm):
    mock_vllm_instance = Mock()
    mock_vllm.LLM.return_value = mock_vllm_instance
    mock_vllm_instance.chat.return_value = [
        _create_vllm_output(["The first time I saw"], "123")
    ]

    engine = VLLMInferenceEngine(_get_default_model_params())
    conversation = Conversation(
        messages=[
            Message(
                content="You're a good assistant!",
                role=Role.SYSTEM,
            ),
            Message(
                content="Hi there",
                role=Role.USER,
            ),
            Message(
                content="Hello again!",
                role=Role.USER,
            ),
        ],
        metadata={"foo": "bar"},
        conversation_id="123",
    )
    expected_result = [
        Conversation(
            messages=[
                *conversation.messages,
                Message(
                    content="The first time I saw",
                    role=Role.ASSISTANT,
                ),
            ],
            metadata={"foo": "bar"},
            conversation_id="123",
        )
    ]
    result = engine.infer([conversation], _get_default_inference_config())
    assert expected_result == result
    mock_vllm_instance.chat.assert_called_once()
    assert isinstance(mock_vllm_instance.chat.call_args_list[0][0][0], list)
    assert mock_vllm_instance.chat.call_args_list[0][0][0] == [
        [
            {
                "content": "You're a good assistant!",
                "role": "system",
            },
            {
                "content": [
                    {
                        "text": "Hi there",
                        "type": "text",
                    },
                    {
                        "text": "Hello again!",
                        "type": "text",
                    },
                ],
                "role": "user",
            },
        ]
    ]


@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_infer_online_multimodal(mock_vllm):
    mock_vllm_instance = Mock()
    mock_vllm.LLM.return_value = mock_vllm_instance
    mock_vllm_instance.chat.return_value = [
        _create_vllm_output(["The first time I saw"], "123")
    ]

    engine = VLLMInferenceEngine(_get_default_model_params())
    conversation = Conversation(
        messages=[
            Message(
                role=Role.USER,
                content=[
                    ContentItem(
                        type=Type.IMAGE_BINARY, binary=_create_test_png_image_bytes()
                    ),
                    ContentItem(type=Type.TEXT, content="Describe this image!"),
                ],
            ),
        ],
        metadata={"foo": "bar"},
        conversation_id="123",
    )
    expected_result = [
        Conversation(
            messages=[
                *conversation.messages,
                Message(
                    content="The first time I saw",
                    role=Role.ASSISTANT,
                ),
            ],
            metadata={"foo": "bar"},
            conversation_id="123",
        )
    ]
    result = engine.infer([conversation], _get_default_inference_config())
    assert expected_result == result
    mock_vllm_instance.chat.assert_called_once()
    assert isinstance(mock_vllm_instance.chat.call_args_list[0][0][0], list)
    assert mock_vllm_instance.chat.call_args_list[0][0][0] == [
        [
            {
                "content": [
                    {
                        "image_url": {"url": _create_test_png_image_base64_str()},
                        "type": "image_url",
                    },
                    {
                        "text": "Describe this image!",
                        "type": "text",
                    },
                ],
                "role": "user",
            }
        ]
    ]


@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_infer_online_lora(mock_vllm, mock_lora_request):
    mock_vllm_instance = Mock()
    mock_vllm.LLM.return_value = mock_vllm_instance
    mock_vllm_instance.chat.return_value = [
        _create_vllm_output(["The first time I saw"], "123")
    ]

    lora_request = LoRARequest(
        lora_name="oumi_lora_adapter",
        lora_int_id=1,
        lora_path="/path/to/adapter",
    )
    mock_lora_request.return_value = lora_request

    with patch("oumi.inference.vllm_inference_engine.get_lora_rank", return_value=32):
        engine = VLLMInferenceEngine(_get_default_model_params(use_lora=True))
    conversation = Conversation(
        messages=[
            Message(
                content="Hello world!",
                role=Role.USER,
            ),
            Message(
                content="Hello again!",
                role=Role.USER,
            ),
        ],
        metadata={"foo": "bar"},
        conversation_id="123",
    )
    expected_result = [
        Conversation(
            messages=[
                *conversation.messages,
                Message(
                    content="The first time I saw",
                    role=Role.ASSISTANT,
                ),
            ],
            metadata={"foo": "bar"},
            conversation_id="123",
        )
    ]
    result = engine.infer([conversation], _get_default_inference_config())
    assert expected_result == result

    mock_lora_request.assert_called_once_with(
        lora_name="oumi_lora_adapter",
        lora_int_id=1,
        lora_path="/path/to/adapter",
    )
    mock_vllm_instance.chat.assert_called_once_with(
        ANY,
        sampling_params=ANY,
        lora_request=lora_request,
        use_tqdm=False,
        chat_template=None,
        chat_template_content_format="auto",
        chat_template_kwargs=None,
    )


@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_infer_online_empty(mock_vllm):
    mock_vllm_instance = Mock()
    mock_vllm.LLM.return_value = mock_vllm_instance
    engine = VLLMInferenceEngine(_get_default_model_params())
    result = engine.infer([], _get_default_inference_config())
    assert [] == result
    mock_vllm_instance.chat.assert_not_called()


@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_infer_online_to_file(mock_vllm):
    mock_vllm_instance = Mock()
    mock_vllm.LLM.return_value = mock_vllm_instance
    mock_vllm_instance.chat.side_effect = [
        [
            _create_vllm_output(["The first time I saw"], "123"),
            _create_vllm_output(["The U.S."], "123"),
        ]
    ]
    with tempfile.TemporaryDirectory() as output_temp_dir:
        engine = VLLMInferenceEngine(_get_default_model_params())
        conversation_1 = Conversation(
            messages=[
                Message(
                    content="Hello world!",
                    role=Role.USER,
                ),
                Message(
                    content="Hello again!",
                    role=Role.USER,
                ),
            ],
            metadata={"foo": "bar"},
            conversation_id="123",
        )
        conversation_2 = Conversation(
            messages=[
                Message(
                    content="Touche!",
                    role=Role.USER,
                ),
            ],
            metadata={"umi": "bar"},
            conversation_id="133",
        )
        expected_result = [
            Conversation(
                messages=[
                    *conversation_1.messages,
                    Message(
                        content="The first time I saw",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"foo": "bar"},
                conversation_id="123",
            ),
            Conversation(
                messages=[
                    *conversation_2.messages,
                    Message(
                        content="The U.S.",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"umi": "bar"},
                conversation_id="133",
            ),
        ]

        output_path = Path(output_temp_dir) / "b" / "output.jsonl"
        inference_config = _get_default_inference_config()
        inference_config.output_path = str(output_path)
        result = engine.infer(
            [conversation_1, conversation_2],
            inference_config,
        )
        assert result == expected_result
        with open(output_path) as f:
            parsed_conversations = []
            for line in f:
                parsed_conversations.append(Conversation.from_json(line))
            assert expected_result == parsed_conversations


@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_infer_from_file(mock_vllm):
    mock_vllm_instance = Mock()
    mock_vllm.LLM.return_value = mock_vllm_instance
    mock_vllm_instance.chat.return_value = [
        _create_vllm_output(["The first time I saw"], "123")
    ]
    with tempfile.TemporaryDirectory() as output_temp_dir:
        engine = VLLMInferenceEngine(_get_default_model_params())
        conversation = Conversation(
            messages=[
                Message(
                    content="Hello world!",
                    role=Role.USER,
                ),
                Message(
                    content="Hello again!",
                    role=Role.USER,
                ),
            ],
            metadata={"foo": "bar"},
            conversation_id="123",
        )
        input_path = Path(output_temp_dir) / "foo" / "input.jsonl"
        _setup_input_conversations(str(input_path), [conversation])
        expected_result = [
            Conversation(
                messages=[
                    *conversation.messages,
                    Message(
                        content="The first time I saw",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"foo": "bar"},
                conversation_id="123",
            )
        ]
        inference_config = _get_default_inference_config()
        inference_config.input_path = str(input_path)
        infer_result = engine.infer(
            inference_config=inference_config,
        )
        assert expected_result == infer_result


@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_infer_from_file_empty(mock_vllm):
    mock_vllm_instance = Mock()
    mock_vllm.LLM.return_value = mock_vllm_instance
    with tempfile.TemporaryDirectory() as output_temp_dir:
        input_path = Path(output_temp_dir) / "foo" / "input.jsonl"
        _setup_input_conversations(str(input_path), [])
        engine = VLLMInferenceEngine(_get_default_model_params())
        inference_config = _get_default_inference_config()
        inference_config.input_path = str(input_path)
        result = engine.infer(inference_config=inference_config)
        assert [] == result
        mock_vllm_instance.chat.assert_not_called()


@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_infer_from_file_to_file(mock_vllm):
    mock_vllm_instance = Mock()
    mock_vllm.LLM.return_value = mock_vllm_instance
    mock_vllm_instance.chat.side_effect = [
        [
            _create_vllm_output(["The first time I saw"], "123"),
            _create_vllm_output(["The U.S."], "123"),
        ]
    ]
    with tempfile.TemporaryDirectory() as output_temp_dir:
        engine = VLLMInferenceEngine(_get_default_model_params())
        conversation_1 = Conversation(
            messages=[
                Message(
                    content="Hello world!",
                    role=Role.USER,
                ),
                Message(
                    content="Hello again!",
                    role=Role.USER,
                ),
            ],
            metadata={"foo": "bar"},
            conversation_id="123",
        )
        conversation_2 = Conversation(
            messages=[
                Message(
                    content="Touche!",
                    role=Role.USER,
                ),
            ],
            metadata={"umi": "bar"},
            conversation_id="133",
        )
        input_path = Path(output_temp_dir) / "foo" / "input.jsonl"
        _setup_input_conversations(str(input_path), [conversation_1, conversation_2])
        expected_result = [
            Conversation(
                messages=[
                    *conversation_1.messages,
                    Message(
                        content="The first time I saw",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"foo": "bar"},
                conversation_id="123",
            ),
            Conversation(
                messages=[
                    *conversation_2.messages,
                    Message(
                        content="The U.S.",
                        role=Role.ASSISTANT,
                    ),
                ],
                metadata={"umi": "bar"},
                conversation_id="133",
            ),
        ]

        output_path = Path(output_temp_dir) / "b" / "output.jsonl"
        inference_config = _get_default_inference_config()
        inference_config.output_path = str(output_path)
        result = engine.infer(
            [conversation_1, conversation_2],
            inference_config,
        )
        assert result == expected_result
        # Ensure the final output is in order.
        with open(output_path) as f:
            parsed_conversations = []
            for line in f:
                parsed_conversations.append(Conversation.from_json(line))
            assert expected_result == parsed_conversations


@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_guided_decoding_json(
    mock_vllm, single_turn_conversation, mock_sampling_params
):
    schema = {
        "type": "object",
        "properties": {"name": {"type": "string"}, "age": {"type": "number"}},
    }
    config = InferenceConfig(
        model=ModelParams(
            model_name="MlpEncoder", tokenizer_name="gpt2", tokenizer_pad_token="<eos>"
        ),
        generation=GenerationParams(guided_decoding=GuidedDecodingParams(json=schema)),
    )

    mock_vllm_instance = Mock()
    mock_vllm.LLM.return_value = mock_vllm_instance
    engine = VLLMInferenceEngine(config.model)
    engine._llm.chat = MagicMock()
    engine._llm.chat.return_value = [
        _create_vllm_output(["The first time I saw"], "123")
    ]
    result = engine._infer([single_turn_conversation], config)

    # Verify SamplingParams was called with guided_decoding
    assert result is not None
    mock_sampling_params.assert_called_once()
    call_kwargs = mock_sampling_params.call_args[1]
    assert "guided_decoding" in call_kwargs
    assert call_kwargs["guided_decoding"].json == schema


@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_guided_decoding_regex(mock_vllm, mock_sampling_params):
    pattern = r"\d{3}-\d{2}-\d{4}"

    config = InferenceConfig(
        model=ModelParams(
            model_name="MlpEncoder", tokenizer_name="gpt2", tokenizer_pad_token="<eos>"
        ),
        generation=GenerationParams(
            guided_decoding=GuidedDecodingParams(regex=pattern)
        ),
    )

    conversation = Conversation(
        messages=[Message(content="Is this a SSN?", role=Role.USER)]
    )

    # Mock the VLLM response
    mock_vllm_instance = Mock()
    mock_vllm.LLM.return_value = mock_vllm_instance
    engine = VLLMInferenceEngine(config.model)

    engine._llm = MagicMock()
    engine._llm.chat.return_value = [MagicMock(outputs=[MagicMock(text="123-45-6789")])]

    result = engine._infer([conversation], config)

    # Verify SamplingParams was called with guided_decoding
    assert result is not None
    mock_sampling_params.assert_called_once()
    call_kwargs = mock_sampling_params.call_args[1]
    assert "guided_decoding" in call_kwargs
    assert call_kwargs["guided_decoding"].regex == pattern


@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_guided_decoding_choice(mock_vllm, mock_sampling_params):
    choices = ["option1", "option2"]
    config = InferenceConfig(
        model=ModelParams(
            model_name="MlpEncoder", tokenizer_name="gpt2", tokenizer_pad_token="<eos>"
        ),
        generation=GenerationParams(
            guided_decoding=GuidedDecodingParams(choice=choices)
        ),
    )

    conversation = Conversation(
        messages=[Message(content="What is your favorite color?", role=Role.USER)]
    )

    # Mock the VLLM response
    mock_vllm_instance = Mock()
    mock_vllm.LLM.return_value = mock_vllm_instance
    engine = VLLMInferenceEngine(config.model)

    engine._llm = MagicMock()
    engine._llm.chat.return_value = [MagicMock(outputs=[MagicMock(text="option1")])]

    result = engine._infer([conversation], config)

    # Verify SamplingParams was called with guided_decoding
    assert result is not None
    mock_sampling_params.assert_called_once()
    call_kwargs = mock_sampling_params.call_args[1]
    assert "guided_decoding" in call_kwargs
    assert call_kwargs["guided_decoding"].choice == choices


@pytest.mark.skipif(vllm_import_failed, reason="vLLM not available")
def test_chat_template_kwargs_enable_thinking_false(mock_vllm):
    convo = Conversation(messages=[Message(content="hi", role=Role.USER)])
    model_params = _get_default_model_params()
    model_params.chat_template_kwargs = {"enable_thinking": False}

    engine = VLLMInferenceEngine(model_params)
    engine._llm = MagicMock()
    engine._llm.chat.return_value = [MagicMock(outputs=[MagicMock(text="response")])]

    inference_config = _get_default_inference_config()
    inference_config.model = model_params

    engine._infer([convo], inference_config=inference_config)

    # Assert the parameter was passed
    call_kwargs = engine._llm.chat.call_args.kwargs
    assert "chat_template_kwargs" in call_kwargs
    assert call_kwargs["chat_template_kwargs"].get("enable_thinking") is False



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Vision-Language SFT collator for conversation-based multimodal training.

This module provides a collator specifically designed for supervised fine-tuning (SFT)
of vision-language models using conversation data.

Unlike VisionLanguageCollatorWithPadding which expects pre-processed features,
this collator works with raw conversation objects and handles the complete feature
generation pipeline.

Example:
    >>> from oumi.builders import build_tokenizer
    >>> from oumi.core.configs import ModelParams
    >>> tokenizer = build_tokenizer(ModelParams(model_name="llava-hf/llava-1.5-7b-hf"))
    >>> collator = VisionLanguageSftCollator(
    ...     tokenizer=tokenizer,
    ...     processor_name="llava-hf/llava-1.5-7b-hf",
    ...     max_length=512,
    ...     truncation=True
    ... )
    >>> # Expects batch items with conversation_json field
    >>> batch = collator([{"conversation_json": conversation1.to_json()}, ...])
"""

from typing import Any, Optional

from oumi.core.feature_generators import (
    FeatureGeneratorOptions,
    VisionLanguageConversationFeatureGenerator,
)
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.core.types import Conversation
from oumi.utils.torch_utils import pad_to_max_dim_and_stack


class VisionLanguageSftCollator:
    """Collator for vision-language SFT that processes conversation data.

    This collator is designed for supervised fine-tuning of vision-language models
    where training data comes in the form of conversations containing both text and
    images. It handles the complete pipeline from raw conversations to model-ready
    tensor batches.

    Key Features:

    - Processes Conversation objects containing text and image data
    - Uses model-specific processors to extract image features
    - Handles tokenization and feature generation in one step
    - Supports various vision-language architectures
    - Manages padding, truncation, and label masking

    The collator expects batch items with a "conversation_json" field containing
    serialized Conversation objects. These conversations can include:

    - Multiple turns of dialogue
    - Image references (paths, URLs, or base64 data)
    - System prompts and user/assistant messages
    """

    def __init__(
        self,
        tokenizer: BaseTokenizer,
        processor_name: str,
        *,
        processor_kwargs: Optional[dict[str, Any]] = None,
        max_length: Optional[int] = None,
        truncation: bool = False,
        truncation_side: str = "right",
        label_ignore_index: Optional[int] = None,
        allow_multi_image_inputs: bool = True,
        trust_remote_code: bool = False,
        train_on_completions_only: bool = False,
        response_template: Optional[str] = None,
        instruction_template: Optional[str] = None,
        process_individually: bool = False,
    ):
        """Initializes the vision-language SFT collator.

        Args:
            tokenizer: The tokenizer for encoding text. Should match the model's
                tokenizer for proper token alignment.

            processor_name: Name or path of the processor to use for feature extraction.
                This should typically match the model name.
                The processor handles image preprocessing and feature extraction.

            processor_kwargs: Optional parameters to pass to the processor constructor.
                These can override default settings or model-specific parameters.

            max_length: Maximum sequence length for padding/truncation. If None,
                sequences are padded to the batch maximum. If specified, sequences
                are padded to this length and may be truncated.

            truncation: Whether to truncate sequences exceeding max_length.
                If False, long sequences are kept intact. Only applies when
                max_length is specified.

            truncation_side: Which side to truncate from ("right" or "left").
                Most models use "right" truncation, but some may require "left"
                for specific architectures or tasks.

            label_ignore_index: Value to use for masking labels in loss computation.

            allow_multi_image_inputs: Whether to support multiple images per
                conversation.
                Set to True for models like MLLaMA that handle multiple images.
                Set to False for models that only support single images per example.

            trust_remote_code: Whether to trust and execute remote code when loading
                the processor. Required for some models (e.g., Qwen2-VL) that use
                custom processing code.

            process_individually: Whether to process each conversation individually
                and then collate features by padding to max dimensions. When True:
                - Each conversation is processed separately through the feature
                    generator
                - Features are padded to the maximum size in the batch
                - Useful for models with variable-sized outputs or heterogeneous data
                - May be less efficient but more flexible than batch processing

                When False (default), conversations are processed as a batch.

            train_on_completions_only: If True, only compute loss on the assistant's
                response tokens. This enables instruction-following training where:

                - **When True**: Only assistant responses contribute to the loss.
                  User instructions, system prompts, and special tokens are masked
                  (ignored) during training. The model learns to generate appropriate
                  responses without being penalized for user input tokens.

                - **When False**: All tokens in the conversation contribute to the loss.
                  This is standard language modeling where the model learns to predict
                  the next token for the entire conversation sequence.

                This parameter is particularly useful for instruction-tuning where you
                want the model to learn response patterns without memorizing prompts.

                **Masking Strategy**: The behavior depends on whether
                instruction_template is provided:

                - **With instruction_template**: Uses multi-turn masking strategy.
                  All user turns are masked, all assistant turns are unmasked.
                  Suitable for multi-turn conversations.

                - **Without instruction_template**: Uses single-turn masking strategy.
                  Only the final assistant response is unmasked, everything else
                  is masked. Suitable for single-turn prompt-response pairs.

            response_template: The template string that marks the beginning of the
                assistant's response. Required if train_on_completions_only is True.

                This should match the exact string that appears in your tokenized
                conversation format to indicate where assistant responses start.

                For example:
                    - Phi-3: "<|assistant|>"
                    - Llama-3: "<|start_header_id|>assistant<|end_header_id|>"
                    - Custom: "Assistant: " or "AI: "

            instruction_template: The template string that marks the beginning of the
                user's instruction.

                For example:
                    - Phi-3: "<|user|>"
                    - Llama-3: "<|start_header_id|>user<|end_header_id|>"
                    - Custom: "User: " or "Human: "
        """
        self._allow_multi_image_inputs = allow_multi_image_inputs
        self._process_individually = process_individually

        if not processor_name:
            raise ValueError("processor_name is required for VisionLanguageSftCollator")

        self._conversation_feature_generator = (
            VisionLanguageConversationFeatureGenerator(
                tokenizer=tokenizer,
                processor_name=processor_name,
                processor_kwargs=processor_kwargs,
                trust_remote_code=trust_remote_code,
                return_tensors="pt",
                truncation=truncation,
                truncation_side=truncation_side,
                max_length=max_length,
                label_ignore_index=label_ignore_index,
                train_on_completions_only=train_on_completions_only,
                response_template=response_template,
                instruction_template=instruction_template,
            )
        )

    def __call__(self, batch) -> dict[str, Any]:
        """Process a batch of conversation data into model-ready features.

        This method converts serialized conversations into the tensor format expected
        by vision-language models. It handles the complete pipeline:
        1. Deserializes conversation JSON strings
        2. Passes conversations to the feature generator
        3. Returns batched tensors ready for training

        Args:
            batch: List of dictionaries, where each dictionary must contain a
                "conversation_json" field with a serialized Conversation object.

                Expected format::

                    [
                        {"conversation_json": '{"messages": [...], "images": [...]}'},
                        {"conversation_json": '{"messages": [...], "images": [...]}'},
                        ...
                    ]

                The conversation JSON should include:
                - messages: List of message dictionaries with role and content
                - images: Optional list of image data (paths, URLs, or base64)

        Returns:
            Dictionary containing all features needed for model training:
                - "input_ids": Token IDs including image placeholders
                - "attention_mask": Attention masks for the input
                - "labels": Target labels with appropriate masking
                - "pixel_values" or model-specific image features
                - Additional model-specific features (cross_attention_mask, etc.)

            The exact keys depend on the model architecture and processor used.

        Raises:
            ValueError: If batch is empty or any item lacks "conversation_json" field.

        Example::

            >>> conversation = Conversation(messages=[
            ...     {"role": "user", "content": "What's in this image?"},
            ...     {"role": "assistant", "content": "I see a cat."}
            ... ], images=["path/to/image.jpg"])
            >>> batch_item = {"conversation_json": conversation.to_json()}
            >>> features = collator([batch_item])
            >>> print(features.keys())
            dict_keys(['input_ids', 'attention_mask', 'labels', 'pixel_values'])

        """
        batch_size = len(batch)
        if batch_size <= 0:
            raise ValueError("Batch is empty")

        conversations: list[Conversation] = []
        for idx in range(batch_size):
            example = batch[idx]
            if "conversation_json" not in example:
                raise ValueError(
                    f"Example doesn't contain 'conversation_json' key. "
                    f"Example: {idx + 1} of {batch_size}. "
                    f"Available keys: {example.keys()}"
                )

            conversation_json = example["conversation_json"]
            conversations.append(Conversation.from_json(conversation_json))
        assert len(conversations) == batch_size

        if self._process_individually:
            individual_results = []
            for conversation in conversations:
                single_result = (
                    self._conversation_feature_generator.transform_conversations(
                        [conversation],
                        FeatureGeneratorOptions(allow_feature_reshape=True),
                    )
                )
                individual_results.append(single_result)

            # Collate features by padding to max dimensions
            result = self._collate_individual_results(individual_results)
        else:
            result = self._conversation_feature_generator.transform_conversations(
                conversations,
                FeatureGeneratorOptions(allow_feature_reshape=False),
            )

        return result

    def _collate_individual_results(
        self, results: list[dict[str, Any]]
    ) -> dict[str, Any]:
        """Collate individually processed results by padding to max dimensions.

        Args:
            results: List of feature dictionaries from individual conversation
                processing

        Returns:
            Collated dictionary with padded tensors

        Raises:
            ValueError: If results have inconsistent keys or non-tensor values
        """
        if not results or len(results) == 0:
            return {}

        # Get keys from first result and verify consistency
        expected_keys = set(results[0].keys())
        for i, result in enumerate(results):
            if set(result.keys()) != expected_keys:
                raise ValueError(
                    f"Inconsistent keys in batch. Expected {expected_keys}, "
                    f"but result {i} has {set(result.keys())}"
                )

        # Collate each feature
        collated = {}
        for key in expected_keys:
            values = [result[key] for result in results]

            # Determine max variable dimensions based on feature type
            # For multi-image models, we may need 2 variable dims (num_images, seq_len)
            max_var_dims = 2 if self._allow_multi_image_inputs else 1

            # Pad and stack tensors
            collated[key] = pad_to_max_dim_and_stack(
                values, max_variable_sized_dims=max_var_dims
            )

        return collated



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit tests for base quantization functionality."""

import pytest  # type: ignore

from oumi.core.configs import ModelParams, QuantizationConfig
from oumi.quantize.base import BaseQuantization, QuantizationResult


class TestQuantization(BaseQuantization):
    """Test implementation of BaseQuantization."""

    supported_methods = ["awq_q4_0"]  # Use a real supported method
    supported_formats = ["safetensors", "pytorch"]

    def quantize(self, config: QuantizationConfig) -> QuantizationResult:
        """Test implementation of quantize."""
        return QuantizationResult(
            quantized_size_bytes=1000,
            output_path="/fake/path",
            quantization_method="awq_q4_0",
            format_type="safetensors",
            additional_info={},
        )

    def raise_if_requirements_not_met(self) -> None:
        """Test implementation."""
        pass


class TestBaseQuantization:
    """Test cases for BaseQuantization functionality."""

    def setup_method(self):
        """Set up test fixtures."""
        self.test_quantizer = TestQuantization()
        self.valid_config = QuantizationConfig(
            model=ModelParams(model_name="test/model"),
            method="awq_q4_0",
            output_path="test_output.safetensors",
            output_format="safetensors",
        )

    def test_supports_method_true(self):
        """Test supports_method returns True for supported method."""
        assert self.test_quantizer.supports_method("awq_q4_0") is True

    def test_supports_method_false(self):
        """Test supports_method returns False for unsupported method."""
        assert self.test_quantizer.supports_method("unsupported") is False

    def test_supports_format_true(self):
        """Test supports_format returns True for supported format."""
        assert self.test_quantizer.supports_format("safetensors") is True
        assert self.test_quantizer.supports_format("pytorch") is True

    def test_supports_format_false(self):
        """Test supports_format returns False for unsupported format."""
        assert self.test_quantizer.supports_format("gguf") is False

    def test_validate_config_valid(self):
        """Test validate_config with valid configuration."""
        # Should not raise any exception
        self.test_quantizer.validate_config(self.valid_config)

    def test_validate_config_invalid_method(self):
        """Test validate_config with invalid method."""
        config = QuantizationConfig(
            model=ModelParams(model_name="test/model"),
            method="bnb_4bit",  # Valid method but not supported by TestQuantization
            output_path="test.safetensors",
            output_format="safetensors",
        )

        with pytest.raises(ValueError, match="not supported by"):
            self.test_quantizer.validate_config(config)

    def test_quantize_implementation(self):
        """Test the quantize method implementation."""
        result = self.test_quantizer.quantize(self.valid_config)

        assert isinstance(result, QuantizationResult)
        assert result.quantized_size_bytes == 1000
        assert result.output_path == "/fake/path"
        assert result.quantization_method == "awq_q4_0"
        assert result.format_type == "safetensors"


class TestQuantizationResult:
    """Test cases for QuantizationResult dataclass."""

    def test_quantization_result_creation(self):
        """Test creating a QuantizationResult instance."""
        result = QuantizationResult(
            quantized_size_bytes=2048,
            output_path="/path/to/model",
            quantization_method="awq_q4_0",
            format_type="safetensors",
            additional_info={"compression_ratio": 0.25},
        )

        assert result.quantized_size_bytes == 2048
        assert result.output_path == "/path/to/model"
        assert result.quantization_method == "awq_q4_0"
        assert result.format_type == "safetensors"
        assert result.additional_info["compression_ratio"] == 0.25

    def test_quantization_result_default_additional_info(self):
        """Test QuantizationResult with default additional_info."""
        result = QuantizationResult(
            quantized_size_bytes=1024,
            output_path="/path",
            quantization_method="method",
            format_type="format",
        )

        assert result.additional_info == {}



import copy
import logging
import sys
import tempfile
from unittest.mock import Mock, patch

import pytest
import typer
from typer.testing import CliRunner

from oumi.cli.cli_utils import CONTEXT_ALLOW_EXTRA_ARGS
from oumi.cli.distributed_run import accelerate, torchrun
from oumi.utils.logging import logger

runner = CliRunner()


#
# Fixtures
#
@pytest.fixture
def app():
    fake_app = typer.Typer()
    fake_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(accelerate)
    fake_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(torchrun)
    yield fake_app


@pytest.fixture
def mock_os():
    with patch("oumi.cli.distributed_run.os") as os_mock:
        yield os_mock


@pytest.fixture
def mock_popen():
    with patch("oumi.cli.distributed_run.Popen") as popen_mock:
        yield popen_mock


@pytest.fixture
def mock_subprocess():
    with patch("oumi.cli.distributed_run.subprocess") as subprocess_mock:
        yield subprocess_mock


@pytest.fixture
def mock_torch():
    torch_mock = Mock()
    with patch.dict("sys.modules", {"torch": torch_mock}):
        yield torch_mock


def test_torchrun_skypilot_single_gpu(
    app,
    mock_os,
    mock_popen,
    monkeypatch,
):
    test_env_vars = {
        "SKYPILOT_NODE_IPS": "mymachine",
        "SKYPILOT_NODE_RANK": 0,
        "SKYPILOT_NUM_GPUS_PER_NODE": 1,
    }
    mock_os.environ.copy.return_value = copy.deepcopy(test_env_vars)

    mock_process = Mock()
    mock_popen.return_value = mock_process
    mock_process.wait.return_value = 0

    monkeypatch.setattr("oumi.cli.distributed_run.sys.stdout", sys.stdout)
    monkeypatch.setattr("oumi.cli.distributed_run.sys.stderr", sys.stderr)

    _ = runner.invoke(
        app,
        [
            "torchrun",
            "-m",
            "oumi",
            "train",
            "--training.max_steps",
            "20",
            "--log-level",
            "ERROR",
        ],
    )

    mock_popen.assert_called_once_with(
        [
            "oumi",
            "train",
            "--training.max_steps",
            "20",
        ],
        env=copy.deepcopy(test_env_vars),
        stdout=sys.stdout,
        stderr=sys.stderr,
        bufsize=1,
        universal_newlines=True,
    )
    assert logger.level == logging.ERROR


def test_torchrun_skypilot_multi_gpu(
    app,
    mock_os,
    mock_popen,
    monkeypatch,
):
    test_env_vars = {
        "SKYPILOT_NODE_IPS": "x111\nx222\nx333\n",
        "SKYPILOT_NODE_RANK": 2,
        "SKYPILOT_NUM_GPUS_PER_NODE": 4,
        # Define the redundant OUMI_ variables to activate consistency checks.
        "OUMI_TOTAL_NUM_GPUS": 12,
        "OUMI_NUM_NODES": 3,
        "OUMI_MASTER_ADDR": "x111",
    }
    mock_os.environ.copy.return_value = copy.deepcopy(test_env_vars)

    mock_process = Mock()
    mock_popen.return_value = mock_process
    mock_process.wait.return_value = 0

    monkeypatch.setattr("oumi.cli.distributed_run.sys.stdout", sys.stdout)
    monkeypatch.setattr("oumi.cli.distributed_run.sys.stderr", sys.stderr)

    _ = runner.invoke(
        app,
        [
            "torchrun",
            "-m",
            "oumi",
            "train",
            "--training.max_steps",
            "20",
            "--log-level",
            "ERROR",
        ],
    )

    mock_popen.assert_called_once_with(
        [
            "torchrun",
            "--nnodes=3",
            "--node-rank=2",
            "--nproc-per-node=4",
            "--master-addr=x111",
            "--master-port=8007",
            "-m",
            "oumi",
            "train",
            "--training.max_steps",
            "20",
        ],
        env=copy.deepcopy(test_env_vars),
        stdout=sys.stdout,
        stderr=sys.stderr,
        bufsize=1,
        universal_newlines=True,
    )
    assert logger.level == logging.ERROR


def test_torchrun_polaris_multi_gpu(
    app,
    mock_os,
    mock_popen,
    monkeypatch,
):
    with tempfile.NamedTemporaryFile("w+t") as file_nodelist:
        # file_nodelist.name
        file_nodelist.writelines(["z111\n", "x222\n", "x333\n"])
        file_nodelist.flush()

        test_env_vars = {
            "PBS_NODEFILE": file_nodelist.name,
            "PMI_RANK": 1,
            "PBS_JOBID": "123456.polaris",
            # Define the redundant OUMI_ variables to activate consistency checks.
            "OUMI_TOTAL_NUM_GPUS": 12,
            "OUMI_NUM_NODES": 3,
            "OUMI_MASTER_ADDR": "z111",
        }
        mock_os.environ.copy.return_value = copy.deepcopy(test_env_vars)

        mock_process = Mock()
        mock_popen.return_value = mock_process
        mock_process.wait.return_value = 0

        monkeypatch.setattr("oumi.cli.distributed_run.sys.stdout", sys.stdout)
        monkeypatch.setattr("oumi.cli.distributed_run.sys.stderr", sys.stderr)

        _ = runner.invoke(
            app,
            [
                "torchrun",
                "-m",
                "oumi",
                "train",
                "--training.max_steps",
                "21",
                "--log-level",
                "DEBUG",
            ],
        )

        mock_popen.assert_called_once()
        mock_popen.assert_called_once_with(
            [
                "torchrun",
                "--nnodes=3",
                "--node-rank=1",
                "--nproc-per-node=4",
                "--master-addr=z111",
                "--master-port=8007",
                "-m",
                "oumi",
                "train",
                "--training.max_steps",
                "21",
            ],
            env=copy.deepcopy(test_env_vars),
            stdout=sys.stdout,
            stderr=sys.stderr,
            bufsize=1,
            universal_newlines=True,
        )
        assert logger.level == logging.DEBUG


def test_torchrun_slurm_multi_gpu(
    app,
    mock_os,
    mock_popen,
    mock_subprocess,
    mock_torch,
    monkeypatch,
):
    test_env_vars = {
        "SLURM_NODELIST": "frontier[04316-04317]",
        "SLURM_NODEID": 1,
        "SLURM_JOBID": "123456.frontier",
        # Define the redundant OUMI_ variables to activate consistency checks.
        "OUMI_TOTAL_NUM_GPUS": 16,
        "OUMI_NUM_NODES": 2,
        "OUMI_MASTER_ADDR": "frontier04316",
    }
    mock_os.environ.copy.return_value = copy.deepcopy(test_env_vars)
    mock_torch.cuda.device_count.return_value = 8

    mock_process = Mock()
    mock_popen.return_value = mock_process
    mock_process.wait.return_value = 0

    mock_subprocess_run = Mock()
    mock_subprocess.run.return_value = mock_subprocess_run
    mock_subprocess_run.stdout = "frontier04316\nfrontier04317"

    monkeypatch.setattr("oumi.cli.distributed_run.sys.stdout", sys.stdout)
    monkeypatch.setattr("oumi.cli.distributed_run.sys.stderr", sys.stderr)

    _ = runner.invoke(
        app,
        [
            "torchrun",
            "-m",
            "oumi",
            "train",
            "--training.max_steps",
            "21",
            "--log-level",
            "DEBUG",
        ],
    )

    mock_popen.assert_called_once()
    mock_popen.assert_called_once_with(
        [
            "torchrun",
            "--nnodes=2",
            "--node-rank=1",
            "--nproc-per-node=8",
            "--master-addr=frontier04316",
            "--master-port=8007",
            "-m",
            "oumi",
            "train",
            "--training.max_steps",
            "21",
        ],
        env=copy.deepcopy(test_env_vars),
        stdout=sys.stdout,
        stderr=sys.stderr,
        bufsize=1,
        universal_newlines=True,
    )
    assert logger.level == logging.DEBUG


def test_accelerate_skypilot_multi_gpu(
    app,
    mock_os,
    mock_popen,
    monkeypatch,
):
    test_env_vars = {
        "SKYPILOT_NODE_IPS": "x111\nx222\nx333\n",
        "SKYPILOT_NODE_RANK": 2,
        "SKYPILOT_NUM_GPUS_PER_NODE": 4,
        # Define the redundant OUMI_ variables to activate consistency checks.
        "OUMI_TOTAL_NUM_GPUS": 12,
        "OUMI_NUM_NODES": 3,
        "OUMI_MASTER_ADDR": "x111",
    }
    mock_os.environ.copy.return_value = copy.deepcopy(test_env_vars)

    mock_process = Mock()
    mock_popen.return_value = mock_process
    mock_process.wait.return_value = 0

    monkeypatch.setattr("oumi.cli.distributed_run.sys.stdout", sys.stdout)
    monkeypatch.setattr("oumi.cli.distributed_run.sys.stderr", sys.stderr)

    _ = runner.invoke(
        app,
        [
            "accelerate",
            "launch",
            "-m",
            "oumi",
            "evaluate",
            "--log-level",
            "DEBUG",
        ],
    )

    mock_popen.assert_called_once_with(
        [
            "accelerate",
            "launch",
            "--num_machines=3",
            "--machine_rank=2",
            "--num_processes=12",
            "--main_process_ip=x111",
            "--main_process_port=8007",
            "-m",
            "oumi",
            "evaluate",
        ],
        env=copy.deepcopy(test_env_vars),
        stdout=sys.stdout,
        stderr=sys.stderr,
        bufsize=1,
        universal_newlines=True,
    )
    assert logger.level == logging.DEBUG


def test_torchrun_localmachine_multi_gpu(
    app,
    mock_os,
    mock_popen,
    mock_torch,
    monkeypatch,
):
    test_env_vars = {
        # No environment vars set
    }
    mock_os.environ.copy.return_value = copy.deepcopy(test_env_vars)
    mock_torch.cuda.device_count.return_value = 8

    mock_process = Mock()
    mock_popen.return_value = mock_process
    mock_process.wait.return_value = 0

    monkeypatch.setattr("oumi.cli.distributed_run.sys.stdout", sys.stdout)
    monkeypatch.setattr("oumi.cli.distributed_run.sys.stderr", sys.stderr)

    _ = runner.invoke(
        app,
        [
            "torchrun",
            "-m",
            "oumi",
            "train",
            "--training.max_steps",
            "20",
            "--log-level",
            "ERROR",
        ],
    )

    mock_popen.assert_called_once_with(
        [
            "torchrun",
            "--nnodes=1",
            "--node-rank=0",
            "--nproc-per-node=8",
            "--master-addr=127.0.0.1",
            "--master-port=8007",
            "-m",
            "oumi",
            "train",
            "--training.max_steps",
            "20",
        ],
        env=copy.deepcopy(test_env_vars),
        stdout=sys.stdout,
        stderr=sys.stderr,
        bufsize=1,
        universal_newlines=True,
    )
    assert logger.level == logging.ERROR


def test_torchrun_localmachine_multi_gpu_masteraddress(
    app,
    mock_os,
    mock_popen,
    mock_torch,
    monkeypatch,
):
    test_env_vars = {"MASTER_ADDRESS": "111.0.0.0", "MASTER_PORT": 1337}
    mock_os.environ.copy.return_value = copy.deepcopy(test_env_vars)
    mock_torch.cuda.device_count.return_value = 8

    mock_process = Mock()
    mock_popen.return_value = mock_process
    mock_process.wait.return_value = 0

    monkeypatch.setattr("oumi.cli.distributed_run.sys.stdout", sys.stdout)
    monkeypatch.setattr("oumi.cli.distributed_run.sys.stderr", sys.stderr)

    _ = runner.invoke(
        app,
        [
            "torchrun",
            "-m",
            "oumi",
            "train",
            "--training.max_steps",
            "20",
            "--log-level",
            "ERROR",
        ],
    )

    mock_popen.assert_called_once_with(
        [
            "torchrun",
            "--nnodes=1",
            "--node-rank=0",
            "--nproc-per-node=8",
            "--master-addr=111.0.0.0",
            "--master-port=1337",
            "-m",
            "oumi",
            "train",
            "--training.max_steps",
            "20",
        ],
        env=copy.deepcopy(test_env_vars),
        stdout=sys.stdout,
        stderr=sys.stderr,
        bufsize=1,
        universal_newlines=True,
    )
    assert logger.level == logging.ERROR



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
import inspect
import logging
import re
from collections.abc import Iterator
from enum import Enum
from io import StringIO
from pathlib import Path
from typing import Any, Optional, TypeVar, Union, cast

from omegaconf import OmegaConf

from oumi.core.configs.params.base_params import BaseParams

T = TypeVar("T", bound="BaseConfig")

_CLI_IGNORED_PREFIXES = ["--local-rank"]

# Set of primitive types that OmegaConf can handle directly
_PRIMITIVE_TYPES = {str, int, float, bool, type(None), bytes, Path, Enum}


def _is_primitive_type(value: Any) -> bool:
    """Check if a value is of a primitive type that OmegaConf can handle."""
    return (
        type(value) in _PRIMITIVE_TYPES
        or isinstance(value, Path)
        or isinstance(value, Enum)
    )


def _handle_non_primitives(config: Any, removed_paths: set, path: str = "") -> Any:
    """Recursively process config object to handle non-primitive values.

    Args:
        config: The config object to process
        removed_paths: Set to track paths of removed non-primitive values
        path: The current path in the config (for logging)

    Returns:
        The processed config with non-primitive values removed
    """
    if _is_primitive_type(config):
        return config

    # Try to convert functions to their source code
    if callable(config):
        try:
            # Lambda functions and built-in functions can't have source extracted
            if hasattr(config, "__name__") and config.__name__ == "<lambda>":
                removed_paths.add(path)
                return None

            source = inspect.getsource(config)
            # Only return source if we successfully got it
            return source
        except (TypeError, OSError):
            # Can't get source for lambdas, built-ins, or C extensions
            removed_paths.add(path)
            return None

    if isinstance(config, list):
        return [
            _handle_non_primitives(item, removed_paths, f"{path}[{i}]")
            for i, item in enumerate(config)
        ]

    # Handle dicts and dataclasses.
    if isinstance(config, dict) or hasattr(config, "__dataclass_fields__"):
        result = {}
        if isinstance(config, dict):
            items = config.items()
        else:  # dataclass
            items = (
                (field_name, getattr(config, field_name))
                for field_name in config.__dataclass_fields__
            )
        for key, value in items:
            # Compose path as per type
            current_path = f"{path}.{key}" if path else key
            if _is_primitive_type(value):
                result[key] = value
            else:
                processed_value = _handle_non_primitives(
                    value, removed_paths, current_path
                )
                if processed_value is not None:
                    result[key] = processed_value
                else:
                    removed_paths.add(current_path)
                    result[key] = None
        return result

    # For any other type, remove it and track the path
    removed_paths.add(path)
    return None


def _filter_ignored_args(arg_list: list[str]) -> list[str]:
    """Filters out ignored CLI arguments."""
    return [
        arg
        for arg in arg_list
        if not any(arg.startswith(prefix) for prefix in _CLI_IGNORED_PREFIXES)
    ]


def _read_config_without_interpolation(config_path: str) -> str:
    """Reads a configuration file without interpolating variables.

    Args:
        config_path: The path to the configuration file.

    Returns:
        str: The stringified configuration.
    """
    with open(config_path) as f:
        stringified_config = f.read()
        pattern = r"(?<!\\)\$\{"  # Matches "${" but not "\${"
        stringified_config = re.sub(pattern, "\\${", stringified_config)
    return stringified_config


@dataclasses.dataclass(eq=False)
class BaseConfig:
    def to_yaml(self, config_path: Union[str, Path, StringIO]) -> None:
        """Saves the configuration to a YAML file.

        Non-primitive values are removed and warnings are logged.

        Args:
            config_path: Path to save the config to
        """
        # Convert dataclass fields to a dictionary first
        config_dict = {}
        for field_name, field_value in self:
            config_dict[field_name] = field_value

        # Process non-primitive values before creating OmegaConf structure
        removed_paths = set()
        processed_config = _handle_non_primitives(
            config_dict, removed_paths=removed_paths
        )

        # Log warnings for removed values
        if removed_paths:
            logger = logging.getLogger(__name__)
            logger.warning(
                "The following non-primitive values were removed from the config "
                "as they cannot be saved to YAML:\n"
                + "\n".join(f"- {path}" for path in sorted(removed_paths))
            )

        OmegaConf.save(config=processed_config, f=config_path)

    @classmethod
    def from_yaml(
        cls: type[T], config_path: Union[str, Path], ignore_interpolation=True
    ) -> T:
        """Loads a configuration from a YAML file.

        Args:
            config_path: The path to the YAML file.
            ignore_interpolation: If True, then any interpolation variables in the
                configuration file will be escaped.

        Returns:
            BaseConfig: The merged configuration object.
        """
        schema = OmegaConf.structured(cls)
        if ignore_interpolation:
            stringified_config = _read_config_without_interpolation(str(config_path))
            file_config = OmegaConf.create(stringified_config)
        else:
            file_config = OmegaConf.load(config_path)
        config = OmegaConf.to_object(OmegaConf.merge(schema, file_config))
        if not isinstance(config, cls):
            raise TypeError(f"config is not {cls}")
        return cast(T, config)

    @classmethod
    def from_str(cls: type[T], config_str: str) -> T:
        """Loads a configuration from a YAML string.

        Args:
            config_str: The YAML string.

        Returns:
            BaseConfig: The configuration object.
        """
        schema = OmegaConf.structured(cls)
        file_config = OmegaConf.create(config_str)
        config = OmegaConf.to_object(OmegaConf.merge(schema, file_config))
        if not isinstance(config, cls):
            raise TypeError(f"config is not {cls}")
        return cast(T, config)

    @classmethod
    def from_yaml_and_arg_list(
        cls: type[T],
        config_path: Optional[str],
        arg_list: list[str],
        logger: Optional[logging.Logger] = None,
        ignore_interpolation=True,
    ) -> T:
        """Loads a configuration from various sources.

        If both YAML and arguments list are provided, then
        parameters specified in `arg_list` have higher precedence.

        Args:
            config_path: The path to the YAML file.
            arg_list: Command line arguments list.
            logger: (optional) Logger.
            ignore_interpolation: If True, then any interpolation variables in the
                configuration file will be escaped.

        Returns:
            BaseConfig: The merged configuration object.
        """
        # Start with an empty typed config. This forces OmegaConf to validate
        # that all other configs are of this structured type as well.
        all_configs = [OmegaConf.structured(cls)]

        # Override with configuration file if provided.
        if config_path is not None:
            if ignore_interpolation:
                stringified_config = _read_config_without_interpolation(config_path)
                all_configs.append(OmegaConf.create(stringified_config))
            else:
                all_configs.append(cls.from_yaml(config_path))

        # Merge base config and config from yaml.
        try:
            # Merge and validate configs
            config = OmegaConf.merge(*all_configs)
        except Exception:
            if logger:
                configs_str = "\n\n".join([f"{config}" for config in all_configs])
                logger.exception(
                    f"Failed to merge {len(all_configs)} Omega configs:\n{configs_str}"
                )
            raise

        # Override config with CLI arguments, in order. The arguments, aka flag names,
        # are dot-separated arguments, ex. `model.model_name`. This also supports
        # arguments indexing into lists, ex. `tasks[0].num_samples` or
        # `tasks.0.num_samples`. This is because the config is already populated and
        # typed, so the indexing is properly interpreted as a list index as opposed to
        # a dictionary key.
        try:
            # Filter out CLI arguments that should be ignored.
            arg_list = _filter_ignored_args(arg_list)
            # Override with CLI arguments.
            config.merge_with_dotlist(arg_list)
        except Exception:
            if logger:
                logger.exception(
                    f"Failed to merge arglist {arg_list} with Omega config:\n{config}"
                )
            raise

        config = OmegaConf.to_object(config)
        if not isinstance(config, cls):
            raise TypeError(f"config {type(config)} is not {type(cls)}")

        return cast(T, config)

    def print_config(self, logger: Optional[logging.Logger] = None) -> None:
        """Prints the configuration in a human-readable format.

        Args:
            logger: Optional logger to use. If None, uses module logger.
        """
        if logger is None:
            logger = logging.getLogger(__name__)

        # Convert dataclass fields to a dictionary first
        config_dict = {}
        for field_name, field_value in self:
            config_dict[field_name] = field_value

        # Process non-primitive values before creating OmegaConf structure
        removed_paths = set()
        processed_config = _handle_non_primitives(
            config_dict, removed_paths=removed_paths
        )

        config_yaml = OmegaConf.to_yaml(processed_config, resolve=True)
        logger.info(f"Configuration:\n{config_yaml}")

    def finalize_and_validate(self) -> None:
        """Finalizes and validates the top level params objects."""
        for _, attr_value in self:
            if isinstance(attr_value, BaseParams):
                attr_value.finalize_and_validate()

        self.__finalize_and_validate__()

    def __finalize_and_validate__(self) -> None:
        """Finalizes and validates the parameters of this object.

        This method can be overridden by subclasses to implement custom
        validation logic.

        In case of validation errors, this method should raise a `ValueError`
        or other appropriate exception.
        """

    def __iter__(self) -> Iterator[tuple[str, Any]]:
        """Returns an iterator over field names and values.

        Note: for an attribute to be a field, it must be declared in the
        dataclass definition and have a type annotation.
        """
        for param in dataclasses.fields(self):
            yield param.name, getattr(self, param.name)

    def __eq__(self, other: object) -> bool:
        """Custom equality comparison that handles callable objects specially."""
        if not isinstance(other, self.__class__):
            return False

        for field_name, field_value in self:
            other_value = getattr(other, field_name)

            # Special handling for callable objects
            if callable(field_value) and callable(other_value):
                if (
                    hasattr(field_value, "__name__")
                    and hasattr(other_value, "__name__")
                    and field_value.__name__ == "<lambda>"
                    and other_value.__name__ == "<lambda>"
                ):
                    # Consider all lambda functions equal for config comparison purposes
                    continue

                # For regular functions, try to compare by source code
                try:
                    field_source = inspect.getsource(field_value).strip()
                    other_source = inspect.getsource(other_value).strip()
                    if field_source != other_source:
                        return False
                except (TypeError, OSError):
                    # If we can't get source, fall back to identity comparison
                    if field_value != other_value:
                        return False
            elif callable(field_value) or callable(other_value):
                # One is callable, the other is not
                return False
            else:
                # Normal comparison for non-callable values
                if field_value != other_value:
                    return False

        return True



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.metadata
import importlib.util
import logging
import os
import platform
import sys
import urllib.parse
from enum import Enum
from pathlib import Path
from typing import Annotated, Optional

import requests
import typer
import yaml
from requests.exceptions import RequestException
from rich.console import Console

from oumi.utils.logging import logger, update_logger_level

CONTEXT_ALLOW_EXTRA_ARGS = {"allow_extra_args": True, "ignore_unknown_options": True}
CONFIG_FLAGS = ["--config", "-c"]
OUMI_FETCH_DIR = "~/.oumi/fetch"
OUMI_GITHUB_RAW = "https://raw.githubusercontent.com/oumi-ai/oumi/main"
_OUMI_PREFIX = "oumi://"

CONSOLE = Console()


def section_header(title, console: Console = CONSOLE):
    """Print a section header with the given title.

    Args:
        title: The title text to display in the header.
        console: The Console object to use for printing.
    """
    console.print(f"\n[blue]{'━' * console.width}[/blue]")
    console.print(f"[yellow]   {title}[/yellow]")
    console.print(f"[blue]{'━' * console.width}[/blue]\n")


def parse_extra_cli_args(ctx: typer.Context) -> list[str]:
    """Parses extra CLI arguments into a list of strings.

    Args:
        ctx: The Typer context object.

    Returns:
        List[str]: The extra CLI arguments
    """
    args = []

    # The following formats are supported:
    # 1. Space separated: "--foo" "2"
    # 2. `=`-separated: "--foo=2"
    try:
        num_args = len(ctx.args)
        idx = 0
        while idx < num_args:
            original_key = ctx.args[idx]
            key = original_key.strip()
            if not key.startswith("--"):
                raise typer.BadParameter(
                    "Extra arguments must start with '--'. "
                    f"Found argument `{original_key}` at position {idx}: `{ctx.args}`"
                )
            # Strip leading "--"

            key = key[2:]
            pos = key.find("=")
            if pos >= 0:
                # '='-separated argument
                value = key[(pos + 1) :].strip()
                key = key[:pos].strip()
                if not key:
                    raise typer.BadParameter(
                        "Empty key name for `=`-separated argument. "
                        f"Found argument `{original_key}` at position {idx}: "
                        f"`{ctx.args}`"
                    )
                idx += 1
            else:
                # Space separated argument
                if idx + 1 >= num_args:
                    raise typer.BadParameter(
                        "Trailing argument has no value assigned. "
                        f"Found argument `{original_key}` at position {idx}: "
                        f"`{ctx.args}`"
                    )
                value = ctx.args[idx + 1].strip()
                idx += 2

            if value.startswith("--"):
                logger.warning(
                    f"Argument value ('{value}') starts with `--`! "
                    f"Key: '{original_key}'"
                )

            cli_arg = f"{key}={value}"
            args.append(cli_arg)
    except ValueError:
        bad_args = " ".join(ctx.args)
        raise typer.BadParameter(
            "Extra arguments must be in `--argname value` pairs. "
            f"Recieved: `{bad_args}`"
        )
    logger.debug(f"\n\nParsed CLI args:\n{args}\n\n")
    return args


def configure_common_env_vars() -> None:
    """Sets common environment variables if needed."""
    if "ACCELERATE_LOG_LEVEL" not in os.environ:
        os.environ["ACCELERATE_LOG_LEVEL"] = "info"
    if "TOKENIZERS_PARALLELISM" not in os.environ:
        os.environ["TOKENIZERS_PARALLELISM"] = "false"


class LogLevel(str, Enum):
    """The available logging levels."""

    DEBUG = logging.getLevelName(logging.DEBUG)
    INFO = logging.getLevelName(logging.INFO)
    WARNING = logging.getLevelName(logging.WARNING)
    ERROR = logging.getLevelName(logging.ERROR)
    CRITICAL = logging.getLevelName(logging.CRITICAL)


def set_log_level(level: Optional[LogLevel]):
    """Sets the logging level for the current command.

    Args:
        level (Optional[LogLevel]): The log level to use.
    """
    if not level:
        return
    uppercase_level = level.upper()
    update_logger_level("oumi", level=uppercase_level)
    CONSOLE.print(f"Set log level to [yellow]{uppercase_level}[/yellow]")


LOG_LEVEL_TYPE = Annotated[
    Optional[LogLevel],
    typer.Option(
        "--log-level",
        "-log",
        help="The logging level for the specified command.",
        show_default=False,
        show_choices=True,
        case_sensitive=False,
        callback=set_log_level,
    ),
]

VERBOSE_TYPE = Annotated[
    bool,
    typer.Option(
        "--verbose",
        "-v",
        help="Enable verbose logging with additional debug information.",
        show_default=True,
    ),
]


def _resolve_oumi_prefix(
    config_path: str, output_dir: Optional[Path] = None
) -> tuple[str, Path]:
    """Resolves oumi:// prefix and determines output directory.

    Args:
        config_path: Path that may contain oumi:// prefix
        output_dir: Optional output directory override

    Returns:
        tuple[str, Path]: (cleaned path, output directory)
    """
    if config_path.lower().startswith(_OUMI_PREFIX):
        config_path = config_path[len(_OUMI_PREFIX) :]

    config_dir = output_dir or os.environ.get("OUMI_DIR") or OUMI_FETCH_DIR
    config_dir = Path(config_dir).expanduser()
    config_dir.mkdir(parents=True, exist_ok=True)

    return config_path, config_dir


def resolve_and_fetch_config(
    config_path: str, output_dir: Optional[Path] = None, force: bool = True
) -> Path:
    """Resolve oumi:// prefix and fetch config if needed.

    Args:
        config_path: Original config path that may contain oumi:// prefix
        output_dir: Optional override for output directory
        force: Whether to overwrite an existing config

    Returns:
        Path: Local path to the config file
    """
    if not config_path.lower().startswith(_OUMI_PREFIX):
        return Path(config_path)

    # Remove oumi:// prefix if present
    new_config_path, config_dir = _resolve_oumi_prefix(config_path, output_dir)

    try:
        # Check destination first
        local_path = (config_dir or Path(OUMI_FETCH_DIR).expanduser()) / new_config_path
        if local_path.exists() and not force:
            msg = f"Config already exists at {local_path}. Use --force to overwrite"
            logger.error(msg)
            raise RuntimeError(msg)

        # Fetch from GitHub
        github_url = f"{OUMI_GITHUB_RAW}/{new_config_path.lstrip('/')}"
        response = requests.get(github_url)
        response.raise_for_status()
        config_content = response.text

        # Validate YAML
        yaml.safe_load(config_content)

        # Save to destination
        if local_path.exists():
            logger.warning(f"Overwriting existing config at {local_path}")
        local_path.parent.mkdir(parents=True, exist_ok=True)

        with open(local_path, "w") as f:
            f.write(config_content)
        logger.info(f"Successfully downloaded config to {local_path}")
    except RequestException as e:
        logger.error(f"Failed to download config from GitHub: {e}")
        raise
    except yaml.YAMLError:
        logger.error("Invalid YAML configuration")
        raise

    return Path(local_path)


def create_github_issue_url(exception: Exception, traceback_str: str) -> str:
    """Create a prefilled GitHub issue URL aligned with the bug report template.

    Args:
        exception: The exception that occurred
        traceback_str: The full traceback string

    Returns:
        str: URL for creating a prefilled GitHub issue
    """
    # 2000 is the max length of a URL to ensure it works with any browser
    MAX_URL_LENGTH = 2000
    base_url = "https://github.com/oumi-ai/oumi/issues/new?"

    python_version = (
        f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
    )
    system_info = f"{platform.system()} {platform.release()}"
    oumi_version = importlib.metadata.version("oumi")
    system_info_content = (
        f"**Please paste the output of `oumi env` here.** "
        f"If you can't run this command, here's basic system info:\n\n"
        f"- Operating system: {system_info}\n"
        f"- Python version: {python_version}\n"
        f"- Oumi version: {oumi_version}"
    )

    title_template = "[Automatic Error Report]: {error_msg}"
    reproduction_template = (
        "Steps to reproduce:\n\n"
        "Command executed: `{command}`\n\n"
        "Stack trace:\n```\n{traceback}\n```"
    )

    exception_str = str(exception)
    command_str = " ".join(sys.argv)

    # Limit the title to 50 chars (unencoded)
    title_error = exception_str[:50]

    command_part = command_str
    error_part = exception_str
    stack_part = traceback_str or ""

    def build_url(command_part: str, error_part: str, stack_part: str) -> str:
        params = {
            "template": "bug-report.yaml",
            "title": title_template.format(error_msg=title_error),
            "what-happened": error_part,
            "reproduction-steps": reproduction_template.format(
                command=command_part, traceback=stack_part
            ),
            "system-info": system_info_content,
        }
        return f"{base_url}{urllib.parse.urlencode(params)}"

    full_url = build_url(command_part, error_part, stack_part)

    # If too long, iteratively shrink fields by encoded length
    while len(full_url) > MAX_URL_LENGTH:
        if stack_part:
            # shrink stack from the front (keep tail)
            cut = max(1, len(stack_part) // 2)
            stack_part = stack_part[cut:]
        elif error_part:
            cut = max(1, len(error_part) // 2)
            error_part = error_part[:-cut]
        elif command_part:
            cut = max(1, len(command_part) // 2)
            command_part = command_part[:-cut]
        else:
            return f"{base_url}template=bug-report.yaml"
        full_url = build_url(command_part, error_part, stack_part)
    return full_url



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.datasets import BasePretrainingDataset
from oumi.core.registry import register_dataset


@register_dataset("wikimedia/wikipedia")
class WikipediaDataset(BasePretrainingDataset):
    """Dataset containing cleaned Wikipedia articles in multiple languages.

    This dataset is built from the Wikipedia dumps (https://dumps.wikimedia.org/)
    with one subset per language, each containing a single train split.
    Each example contains the content of one full Wikipedia article
    with cleaning to strip markdown and unwanted sections (references, etc.).

    Data Fields:
        id (str): ID of the article.
        url (str): URL of the article.
        title (str): Title of the article.
        text (str): Text content of the article.

    Note:
        All configurations contain a single 'train' split.

    Languages:
        The dataset supports numerous languages. For a full list, see:
        https://meta.wikimedia.org/wiki/List_of_Wikipedias

    Note:
        The dataset is licensed under the GNU Free Documentation License (GFDL) and
        the Creative Commons Attribution-Share-Alike 3.0 License.

    See Also:
        - Homepage: https://dumps.wikimedia.org
        - Hugging Face Hub: https://huggingface.co/datasets/wikimedia/wikipedia
    """

    default_dataset = "wikimedia/wikipedia"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from oumi.core.configs import JobConfig
from oumi.core.launcher import BaseCloud, BaseCluster, JobStatus
from oumi.core.registry import register_cloud_builder
from oumi.launcher.clients.local_client import LocalClient
from oumi.launcher.clusters.local_cluster import LocalCluster


class LocalCloud(BaseCloud):
    """A resource pool for managing Local jobs.

    It is important to note that a single LocalCluster can only run one job at a time.
    Running multiple GPU jobs simultaneously on separate LocalClusters is not
    encouraged.
    """

    # The default cluster name. Used when no cluster name is provided.
    _DEFAULT_CLUSTER = "local"

    def __init__(self):
        """Initializes a new instance of the LocalCloud class."""
        # A mapping from cluster names to Local Cluster instances.
        self._clusters = {}

    def _get_or_create_cluster(self, name: str) -> LocalCluster:
        """Gets the cluster with the specified name, or creates one if it doesn't exist.

        Args:
            name: The name of the cluster.

        Returns:
            LocalCluster: The cluster instance.
        """
        if name not in self._clusters:
            self._clusters[name] = LocalCluster(name, LocalClient())
        return self._clusters[name]

    def up_cluster(self, job: JobConfig, name: Optional[str], **kwargs) -> JobStatus:
        """Creates a cluster and starts the provided Job."""
        # The default cluster.
        cluster_name = name or self._DEFAULT_CLUSTER
        cluster = self._get_or_create_cluster(cluster_name)
        job_status = cluster.run_job(job)
        if not job_status:
            raise RuntimeError("Failed to start job.")
        return job_status

    def get_cluster(self, name) -> Optional[BaseCluster]:
        """Gets the cluster with the specified name, or None if not found."""
        clusters = self.list_clusters()
        for cluster in clusters:
            if cluster.name() == name:
                return cluster
        return None

    def list_clusters(self) -> list[BaseCluster]:
        """Lists the active clusters on this cloud."""
        return list(self._clusters.values())


@register_cloud_builder("local")
def Local_cloud_builder() -> LocalCloud:
    """Builds a LocalCloud instance."""
    return LocalCloud()



"""Test execution of notebooks."""

from pathlib import Path

import nbformat
import pytest
from nbconvert.preprocessors import ExecutePreprocessor

from tests import get_notebooks_dir

# Notebooks whose tests we should skip, as they can't be run in an automated fashion.
# Reasons include that they're too slow, they start remote jobs, etc.
# An error will be thrown if a notebook in this set isn't found.
_NOTEBOOKS_TO_SKIP = {
    # Requires calling GPT-4 API
    "Oumi - Evaluation with AlpacaEval 2.0.ipynb",
    # Requires calling GPT-4 API
    "Oumi - Evaluation with MT Bench.ipynb",
    # Requires creating a remote instance
    "Oumi - Running Jobs Remotely.ipynb",
    # Requires creating a remote instance
    "Oumi - Using vLLM Engine for Inference.ipynb",
}


def get_notebooks():
    """Get all notebooks in the notebooks directory."""
    notebooks_dir = get_notebooks_dir()
    notebooks_to_skip = _NOTEBOOKS_TO_SKIP.copy()
    notebooks_to_test = []
    for notebook_path in notebooks_dir.glob("*.ipynb"):
        if notebook_path.name in notebooks_to_skip:
            notebooks_to_skip.remove(notebook_path.name)
        else:
            notebooks_to_test.append(notebook_path)
    if notebooks_to_skip:
        raise ValueError(f"Notebooks to skip not found: {notebooks_to_skip}")
    return notebooks_to_test


@pytest.mark.parametrize("notebook_path", get_notebooks(), ids=lambda x: x.name)
def test_notebook_execution(notebook_path: Path):
    """Test that a notebook executes successfully."""
    with open(notebook_path) as f:
        nb = nbformat.read(f, as_version=4)

    ep = ExecutePreprocessor(timeout=600)  # 10 minute timeout

    try:
        ep.preprocess(nb, {"metadata": {"path": notebook_path.parent}})
    except Exception as e:
        pytest.fail(f"Error executing notebook {notebook_path.name}: {str(e)}")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple Bitnet model saving callback."""

from pathlib import Path
from typing import Optional, Union

import transformers
from transformers.trainer_utils import PREFIX_CHECKPOINT_DIR

from oumi.core.callbacks.base_trainer_callback import BaseTrainerCallback
from oumi.core.configs import TrainingParams

# Import `onebitllms` utils methods
try:
    import onebitllms  # type: ignore
    from onebitllms import quantize_to_1bit  # type: ignore
except ImportError:
    onebitllms = None


class BitNetCallback(BaseTrainerCallback):
    """BitNet model saving callback.

    Simple callback that saves the model into BitNet quantized
    format during training.
    """

    def on_save(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Saving callback.

        Gets triggered at each saving step to quantize trained models
        in 1bit precision.
        """
        if onebitllms is None:
            raise ValueError(
                """You need `onebitllms` to be installed in order to save
                correctly BitNet models - `pip install onebitllms`"""
            )

        output_dir = Path(args.output_dir)  # type: ignore
        quantized_subdir = Path(
            f"{PREFIX_CHECKPOINT_DIR}-{state.global_step}-quantized"  # type: ignore
        )
        output_subdir = Path(f"{PREFIX_CHECKPOINT_DIR}-{state.global_step}")  # type: ignore

        checkpoint_folder = output_dir / output_subdir
        quantized_checkpoint_folder = output_dir / quantized_subdir

        quantize_to_1bit(str(checkpoint_folder), str(quantized_checkpoint_folder))



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

from typing_extensions import override

from oumi.core.configs.inference_config import InferenceConfig
from oumi.core.configs.judge_config import JudgeConfig
from oumi.core.configs.params.judge_params import (
    JudgeOutputType,
    JudgeParams,
    JudgeResponseFormat,
)
from oumi.core.inference import BaseInferenceEngine
from oumi.judges.base_judge import (
    BaseJudge,
    JudgeOutputField,
)

# Expected field/key names in the judge's output.
EXPLANATION_KEY = "explanation"
JUDGMENT_KEY = "judgment"

# Judgment options: describing to the judge how to format its judgment.
JUDGMENT_OPTIONS_BOOL = "Your judgment should be a single word: 'Yes' or 'No'"
JUDGMENT_OPTIONS_INT = "Your judgment should be an integer value"
JUDGMENT_OPTIONS_FLOAT = "Your judgment should be a float value"
JUDGMENT_OPTIONS_ENUM_PREFIX = "Your judgment should be one of the following options: "
JUDGMENT_OPTIONS_TEXT = "Your judgment should be provided in the form of free text"

# Prompt suffix: describing to the judge how to format its response (XML, JSON, or RAW).
XML_SUFFIX = (
    "\n\nProvide your response in XML format only. Include your judgment enclosed "
    "within <{judgment_key}> and </{judgment_key}> tags. {judgment_options}Do not  "
    "include any text outside the XML. Ensure that all tags are properly closed and "
    "that the XML is well-formed."
)
XML_SUFFIX_WITH_EXPLANATION = (
    "\n\nProvide your response in XML format only. Begin with an explanation "
    "justifying your judgment, enclosed within <{explanation_key}> and "
    "</{explanation_key}> tags. Follow this with your judgment, enclosed within "
    "<{judgment_key}> and </{judgment_key}> tags. {judgment_options}Do not include any "
    "text outside the XML. Ensure that all tags are properly closed and that the XML "
    "is well-formed."
)
JSON_SUFFIX = (
    "\n\nProvide your response in JSON format only. Include your judgment as the value "
    "of a single key named '{judgment_key}'. {judgment_options}Do not include any "
    "text outside the JSON. Ensure the JSON is properly formatted and valid."
)
JSON_SUFFIX_WITH_EXPLANATION = (
    "\n\nProvide your response in JSON format only. Begin with an explanation "
    "justifying your judgment, using the key '{explanation_key}'. Then include your "
    "judgment using the key '{judgment_key}'. {judgment_options}Do not include any "
    "text outside the JSON. Ensure the JSON is properly formatted and valid."
)
RAW_SUFFIX_WITH_EXPLANATION = (
    "\n\nExplain your reasoning before providing your judgment."
)


class SimpleJudge(BaseJudge):
    """Judge class for evaluating outputs based on a given configuration."""

    def __init__(
        self,
        judge_config: Union[JudgeConfig, str],
    ):
        """Initialize the Judge.

        Args:
            judge_config: JudgeConfig object or a path to a judge configuration file.
                Contains both judge parameters and inference configuration.
        """
        if isinstance(judge_config, str):
            judge_config = JudgeConfig.from_path(judge_config)

        self._judge_params = judge_config.judge_params
        self._judge_params.replace_template_variables()
        self._inference_config = judge_config.inference_config

        # Create output fields based on judge configuration
        output_fields = []
        if self._judge_params.include_explanation:
            output_fields.append(self._create_explanation_output_field())
        output_fields.append(self._create_judgment_output_field(self._judge_params))

        # Generate an inference engine from inference config
        inference_engine = self._create_inference_engine(self._inference_config)

        # Append format suffix to system instruction if it exists
        system_instruction = self._judge_params.system_instruction
        if system_instruction:
            system_instruction = f"{system_instruction}{self._get_format_suffix()}"

        # Get set of prompt template placeholders
        prompt_template_placeholders_set = (
            set(self._judge_params.prompt_template_placeholders)
            if self._judge_params.prompt_template_placeholders
            else self._judge_params.get_placeholders()
        )

        super().__init__(
            prompt_template=self._judge_params.prompt_template,
            prompt_template_placeholders=prompt_template_placeholders_set,
            system_instruction=system_instruction,
            example_field_values=self._judge_params.examples,
            response_format=self._judge_params.response_format,
            output_fields=output_fields,
            inference_engine=inference_engine,
        )

    @override
    def _build_judgment_prompt(self, judge_input: dict[str, str]) -> str:
        """Generate judge prompts using the template."""
        prompt_content = super()._build_judgment_prompt(judge_input)

        # Only append format suffix to judgment prompt if no system instruction exists
        # (otherwise it was already appended to system instruction in __init__)
        if not self._judge_params.system_instruction:
            prompt_content += self._get_format_suffix()

        return prompt_content

    def _get_format_suffix(self) -> str:
        """Get the appropriate format suffix based on response format and explanation.

        Returns:
            Format-specific instruction suffix to append to prompts
        """
        response_format = self._judge_params.response_format
        include_explanation = self._judge_params.include_explanation

        # Describe the expected judgment options to the judge
        if (
            self._judge_params.judgment_scores
            and len(self._judge_params.judgment_scores) > 1
        ):
            choices = [f"'{c}'" for c in self._judge_params.judgment_scores.keys()]
            choices_str = ", ".join(choices)
            judgment_options = f"{JUDGMENT_OPTIONS_ENUM_PREFIX}{choices_str}. "
        elif self._judge_params.judgment_type == JudgeOutputType.BOOL:
            judgment_options = f"{JUDGMENT_OPTIONS_BOOL}. "
        elif self._judge_params.judgment_type == JudgeOutputType.FLOAT:
            judgment_options = f"{JUDGMENT_OPTIONS_FLOAT}. "
        elif self._judge_params.judgment_type == JudgeOutputType.INT:
            judgment_options = f"{JUDGMENT_OPTIONS_INT}. "
        elif self._judge_params.judgment_type == JudgeOutputType.TEXT:
            judgment_options = f"{JUDGMENT_OPTIONS_TEXT}. "
        else:
            judgment_options = ""

        # Describe the expected response format to the judge
        if response_format == JudgeResponseFormat.XML:
            suffix = XML_SUFFIX_WITH_EXPLANATION if include_explanation else XML_SUFFIX
        elif response_format == JudgeResponseFormat.JSON:
            suffix = (
                JSON_SUFFIX_WITH_EXPLANATION if include_explanation else JSON_SUFFIX
            )
        elif response_format == JudgeResponseFormat.RAW:
            suffix = RAW_SUFFIX_WITH_EXPLANATION if include_explanation else ""
        else:
            suffix = ""

        return suffix.format(
            judgment_key=JUDGMENT_KEY,
            explanation_key=EXPLANATION_KEY,
            judgment_options=judgment_options,
        )

    def _create_judgment_output_field(self, params: JudgeParams) -> JudgeOutputField:
        """Create the main judgment output field."""
        return JudgeOutputField(
            field_key=JUDGMENT_KEY,
            field_type=params.judgment_type,
            field_scores=params.judgment_scores,
        )

    def _create_explanation_output_field(self) -> JudgeOutputField:
        """Create the explanation output field."""
        return JudgeOutputField(
            field_key=EXPLANATION_KEY,
            field_type=JudgeOutputType.TEXT,
            field_scores=None,
        )

    def _create_inference_engine(
        self, inference_config: InferenceConfig
    ) -> BaseInferenceEngine:
        """Create the inference engine based on the provided configuration."""
        from oumi.builders.inference_engines import build_inference_engine

        if inference_config.engine is None:
            raise ValueError("Inference engine not specified in the configuration.")
        elif inference_config.input_path or inference_config.output_path:
            raise ValueError(
                "Input and output paths are not supported in inference_config, when "
                "instantiating the SimpleJudge. Please set both to None."
            )

        return build_inference_engine(
            engine_type=inference_config.engine,
            model_params=inference_config.model,
            remote_params=inference_config.remote_params,
            generation_params=inference_config.generation,
        )



from unittest.mock import Mock

import numpy as np
import pytest

from oumi.core.tokenizers.utils import (
    find_all_sequences,
    mask_labels_for_completions_only,
    mask_labels_without_user_template,
)


# Find All Sequences Tests
@pytest.mark.parametrize(
    "sequence,target,expected",
    [
        ([1, 2, 3, 4, 5], [3, 4], [4]),
        ([1, 2, 3, 4, 5], [1, 2], [2]),
        ([1, 2, 3, 4, 5], [4, 5], [5]),
        ([1, 2, 3, 4, 5], [6, 7], []),
        ([1, 2, 3, 4, 5], [1, 2, 3, 4, 5], [5]),
        ([], [1], []),
        ([1, 2, 3, 4, 3, 4, 5], [3, 4], [4, 6]),  # Multiple occurrences
    ],
)
def test_find_all_sequences(sequence, target, expected):
    arr = np.array(sequence)
    result = find_all_sequences(arr, target)
    assert result == expected


# Mask Labels Tests - Basic scenarios
def test_basic_masking_no_user_template():
    """Test basic masking without user template (last assistant turn only strategy)."""
    labels = np.array([1, 2, 3, 4, 5, 6, 7, 8])
    response_tokens = [3, 4]

    mask_labels_without_user_template(labels, response_tokens)

    # Should mask everything except the last assistant response content: [5, 6, 7, 8]
    # Template [3, 4] is masked, only content after it [5, 6, 7, 8] is kept
    expected = np.array([-100, -100, -100, -100, 5, 6, 7, 8])
    np.testing.assert_array_equal(labels, expected)


def test_masking_with_user_template():
    """Test masking with both user and assistant templates."""
    # Conversation: User: [200, 201, 10] Assistant: [100, 101, 20, 21]
    # User: [200, 201, 30] Assistant: [100, 101, 40, 41]
    labels = np.array([200, 201, 10, 100, 101, 20, 21, 200, 201, 30, 100, 101, 40, 41])
    response_tokens = [100, 101]  # Assistant template
    instruction_tokens = [200, 201]  # User template

    mask_labels_for_completions_only(labels, response_tokens, instruction_tokens)

    # Should mask everything except assistant responses: [20, 21] and [40, 41]
    # (templates are masked)
    expected = np.array(
        [-100, -100, -100, -100, -100, 20, 21, -100, -100, -100, -100, -100, 40, 41]
    )
    np.testing.assert_array_equal(labels, expected)


def test_no_response_template_found():
    """Test when response template is not found."""
    labels = np.array([1, 2, 3, 4, 5])
    response_tokens = [9, 10]

    mask_labels_without_user_template(labels, response_tokens)

    # Should mask everything
    expected = np.array([-100, -100, -100, -100, -100])
    np.testing.assert_array_equal(labels, expected)


def test_custom_ignore_index():
    """Test using custom ignore index."""
    labels = np.array([1, 2, 3, 4, 5])
    response_tokens = [2, 3]

    mask_labels_without_user_template(labels, response_tokens, ignore_index=-999)

    # Should mask [1, 2, 3] (including template) and keep [4, 5]
    expected = np.array([-999, -999, -999, 4, 5])
    np.testing.assert_array_equal(labels, expected)


def test_response_at_start():
    """Test when response template is at the beginning."""
    labels = np.array([1, 2, 3, 4, 5])
    response_tokens = [1, 2]

    mask_labels_without_user_template(labels, response_tokens)

    # Should mask the template [1, 2] and keep only the last response content [3, 4, 5]
    expected = np.array([-100, -100, 3, 4, 5])
    np.testing.assert_array_equal(labels, expected)


def test_multiple_responses_no_user_template():
    """Test multiple response templates without user template."""
    labels = np.array([1, 2, 3, 4, 5, 3, 4, 6, 7])
    response_tokens = [3, 4]

    mask_labels_without_user_template(labels, response_tokens)

    # Should mask everything except the last assistant response: [6, 7]
    # First response [5] should be masked since we only train on the last assistant turn
    expected = np.array([-100, -100, -100, -100, -100, -100, -100, 6, 7])
    np.testing.assert_array_equal(labels, expected)


def test_single_turn_conversation():
    """Test single-turn conversation with user and assistant templates."""
    # User: [200, 201, 10, 11] Assistant: [100, 101, 20, 21, 22]
    labels = np.array([200, 201, 10, 11, 100, 101, 20, 21, 22])
    response_tokens = [100, 101]
    instruction_tokens = [200, 201]

    mask_labels_for_completions_only(labels, response_tokens, instruction_tokens)

    # Should mask everything except assistant response content: [20, 21, 22]
    # (template [100, 101] is masked)
    expected = np.array([-100, -100, -100, -100, -100, -100, 20, 21, 22])
    np.testing.assert_array_equal(labels, expected)


# Simpler Feature Generator Fixture
@pytest.fixture
def simple_feature_generator():
    """Simple mock feature generator for testing completion-only masking."""
    fg = Mock()
    fg._response_token_ids = [100, 101]  # "Assistant:"
    fg._instruction_token_ids = [200, 201]  # "User:"

    # Mock the special tokens
    special_tokens = Mock()
    special_tokens.label_ignore_index = -100
    fg._special_tokens = special_tokens

    # Import the actual masking methods we need to test
    from oumi.core.feature_generators.vision_language_conversation_feature_generator import (  # noqa: E501
        VisionLanguageConversationFeatureGenerator,
    )

    # Bind the actual methods to our mock
    fg_class = VisionLanguageConversationFeatureGenerator
    fg._mask_single_conversation = fg_class._mask_single_conversation.__get__(fg)
    fg._apply_completion_only_masking = fg_class._apply_completion_only_masking.__get__(
        fg
    )

    return fg


def test_find_all_template_positions(simple_feature_generator):
    """Test finding all template positions in sequence."""
    from oumi.core.tokenizers.utils import find_all_sequences

    input_ids = np.array([1, 100, 101, 2, 3, 100, 101, 4, 5])
    positions = find_all_sequences(input_ids, [100, 101])
    assert positions == [3, 7]  # Positions after the template


def test_mask_single_conversation_with_user_template(simple_feature_generator):
    """Test masking single conversation with user template."""
    # User: [200, 201, 10] Assistant: [100, 101, 20]
    # User: [200, 201, 30] Assistant: [100, 101, 40]
    input_ids = np.array([200, 201, 10, 100, 101, 20, 200, 201, 30, 100, 101, 40])
    labels = np.array([200, 201, 10, 100, 101, 20, 200, 201, 30, 100, 101, 40])

    simple_feature_generator._mask_single_conversation(labels, input_ids)

    # Should mask everything except assistant response content: 20 and 40
    # (templates are masked)
    expected = np.array(
        [-100, -100, -100, -100, -100, 20, -100, -100, -100, -100, -100, 40]
    )
    np.testing.assert_array_equal(labels, expected)


def test_mask_single_conversation_no_user_template(simple_feature_generator):
    """Test masking single conversation without user template."""
    # Remove user template info
    simple_feature_generator._instruction_token_ids = None

    input_ids = np.array([1, 2, 100, 101, 3, 4, 5])
    labels = np.array([1, 2, 100, 101, 3, 4, 5])

    simple_feature_generator._mask_single_conversation(labels, input_ids)

    # Should mask everything including template and keep only response content
    expected = np.array([-100, -100, -100, -100, 3, 4, 5])
    np.testing.assert_array_equal(labels, expected)


def test_apply_completion_only_masking_list(simple_feature_generator):
    """Test applying completion-only masking to list inputs."""
    inputs = {
        "labels": [[1, 2, 100, 101, 3, 4, 5], [10, 11, 100, 101, 20, 30, 40]],
        "input_ids": [[1, 2, 100, 101, 3, 4, 5], [10, 11, 100, 101, 20, 30, 40]],
    }

    simple_feature_generator._apply_completion_only_masking(inputs)

    # Should mask everything including templates and keep only response content
    expected = [[-100, -100, -100, -100, 3, 4, 5], [-100, -100, -100, -100, 20, 30, 40]]
    assert inputs["labels"] == expected


def test_apply_completion_only_masking_numpy(simple_feature_generator):
    """Test applying completion-only masking to numpy inputs."""
    inputs = {
        "labels": np.array([[1, 2, 100, 101, 3, 4, 5]]),
        "input_ids": np.array([[1, 2, 100, 101, 3, 4, 5]]),
    }

    simple_feature_generator._apply_completion_only_masking(inputs)

    # Should mask everything including template and keep only response content
    expected = np.array([[-100, -100, -100, -100, 3, 4, 5]])
    np.testing.assert_array_equal(inputs["labels"], expected)


# Edge Case Tests
def test_empty_labels():
    """Test with empty labels array."""
    labels = np.array([])
    mask_labels_without_user_template(labels, [1, 2])
    assert len(labels) == 0


def test_response_template_longer_than_sequence():
    """Test when response template is longer than the entire sequence."""
    labels = np.array([1, 2])
    response_tokens = [1, 2, 3, 4, 5]

    mask_labels_without_user_template(labels, response_tokens)

    # Should mask everything since template not found
    expected = np.array([-100, -100])
    np.testing.assert_array_equal(labels, expected)


def test_no_assistant_response_with_user_template():
    """Test conversation with user template but no assistant response."""
    labels = np.array([200, 201, 10, 11, 12])  # Only user message
    response_tokens = [100, 101]  # Assistant template
    instruction_tokens = [200, 201]  # User template

    mask_labels_for_completions_only(labels, response_tokens, instruction_tokens)

    # Should mask everything since no assistant response found
    expected = np.array([-100, -100, -100, -100, -100])
    np.testing.assert_array_equal(labels, expected)


def test_assistant_response_at_end():
    """Test when assistant response is at the very end."""
    labels = np.array([200, 201, 10, 100, 101])
    response_tokens = [100, 101]
    instruction_tokens = [200, 201]

    mask_labels_for_completions_only(labels, response_tokens, instruction_tokens)

    # Should mask everything including the assistant template
    # (no content after template)
    expected = np.array([-100, -100, -100, -100, -100])
    np.testing.assert_array_equal(labels, expected)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Supervised fine-tuning datasets module."""

from oumi.datasets.sft.alpaca import AlpacaDataset
from oumi.datasets.sft.aya import AyaDataset
from oumi.datasets.sft.chatqa import ChatqaDataset, ChatqaTatqaDataset
from oumi.datasets.sft.chatrag_bench import ChatRAGBenchDataset
from oumi.datasets.sft.coalm import CoALMDataset
from oumi.datasets.sft.dolly import ArgillaDollyDataset
from oumi.datasets.sft.huggingface import HuggingFaceDataset
from oumi.datasets.sft.magpie import ArgillaMagpieUltraDataset, MagpieProDataset
from oumi.datasets.sft.openo1_sft import OpenO1SFTDataset
from oumi.datasets.sft.prompt_response import PromptResponseDataset
from oumi.datasets.sft.sft_jsonlines import TextSftJsonLinesDataset
from oumi.datasets.sft.tulu3_sft_mixture import Tulu3MixtureDataset
from oumi.datasets.sft.ultrachat import UltrachatH4Dataset
from oumi.datasets.sft.wildchat import WildChatDataset

__all__ = [
    "AlpacaDataset",
    "ArgillaDollyDataset",
    "ArgillaMagpieUltraDataset",
    "AyaDataset",
    "ChatqaDataset",
    "ChatqaTatqaDataset",
    "ChatRAGBenchDataset",
    "CoALMDataset",
    "HuggingFaceDataset",
    "MagpieProDataset",
    "OpenO1SFTDataset",
    "PromptResponseDataset",
    "TextSftJsonLinesDataset",
    "Tulu3MixtureDataset",
    "UltrachatH4Dataset",
    "WildChatDataset",
]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Models module for the Oumi (Open Universal Machine Intelligence) library.

This module provides various model implementations for use in the Oumi framework.
These models are designed for different machine learning tasks and can be used
with the datasets and training pipelines provided by Oumi.

Available models:
    - :py:class:`~oumi.models.mlp.MLPEncoder`: A Multi-Layer Perceptron (MLP)
        encoder model.
    - :py:class:`~oumi.models.cnn_classifier.CNNClassifier`: A simple ConvNet for
        image classification e.g., can be used for MNIST digits classification.

Each model is implemented as a separate class, inheriting from appropriate base classes
in the Oumi framework.

Example:
    >>> from oumi.models import MLPEncoder
    >>> encoder = MLPEncoder(input_dim=784, hidden_dim=256, output_dim=10)
    >>> output = encoder(input_data) # doctest: +SKIP

Note:
    For detailed information on each model, please refer to their respective
    class documentation.
"""

from oumi.models.cnn_classifier import CNNClassifier
from oumi.models.mlp import MLPEncoder

__all__ = ["MLPEncoder", "CNNClassifier"]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import importlib.util
import os
import sys
from collections import namedtuple
from enum import Enum, auto
from pathlib import Path
from typing import Any, Callable, Optional

from oumi.utils.logging import logger


class RegistryType(Enum):
    CLOUD = auto()
    DATASET = auto()
    METRICS_FUNCTION = auto()
    REWARD_FUNCTION = auto()
    ROLLOUT_FUNCTION = auto()
    MODEL_CONFIG = auto()
    MODEL = auto()
    JUDGE_CONFIG = auto()
    EVALUATION_FUNCTION = auto()
    SAMPLE_ANALYZER = auto()


class RegistryKey(namedtuple("RegistryKey", ["name", "registry_type"])):
    def __new__(cls, name: str, registry_type: RegistryType):
        """Create a new RegistryKey instance.

        Args:
            name: The name of the registry key.
            registry_type: The type of the registry.

        Returns:
            A new RegistryKey instance with lowercase name.
        """
        return super().__new__(cls, name.lower(), registry_type)


def _load_user_requirements(requirements_file: str):
    """Loads user-defined requirements from a file."""
    logger.info(f"Loading user-defined registry from: {requirements_file}")
    logger.info(
        "This value can be set using the OUMI_EXTRA_DEPS_FILE environment variable."
    )
    requirements_path = Path(requirements_file)
    if not requirements_path.exists():
        logger.error(f"OUMI_EXTRA_DEPS_FILE file not found: {requirements_file}")
        raise FileNotFoundError(
            f"OUMI_EXTRA_DEPS_FILE file not found: {requirements_file}"
        )
    with open(requirements_path) as f:
        import_count = 0
        for idx, line in enumerate(f):
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            import_count += 1
            import_path = Path(line)
            logger.info(f"Loading user-defined registry module: {import_path}")
            mod_name = import_path.stem
            sys.path.append(str(import_path.parent))
            try:
                importlib.import_module(mod_name)
            except Exception as e:
                logger.error(
                    "Failed to load a user-defined module in "
                    f"OUMI_EXTRA_DEPS_FILE: {line}"
                )
                raise ImportError(f"Failed to load user-defined module: {line}") from e
        logger.info(f"Loaded {import_count} user-defined registry modules.")


def _register_dependencies(cls_function):
    """Decorator to ensure core dependencies are added to the Registry."""

    @functools.wraps(cls_function)
    def wrapper(self, *args, **kwargs):
        if not self._initialized:
            # Immediately set the initialized flag to avoid infinite recursion.
            self._initialized = True
            # Import all core dependencies.
            import oumi.core.analyze  # noqa: F401
            import oumi.datasets  # noqa: F401
            import oumi.launcher  # noqa: F401
            import oumi.models  # noqa: F401

            # Import user-defined dependencies.
            user_req_file = os.environ.get("OUMI_EXTRA_DEPS_FILE", None)
            if user_req_file:
                _load_user_requirements(user_req_file)

        return cls_function(self, *args, **kwargs)

    return wrapper


class Registry:
    _initialized: bool = False

    def __init__(self):
        """Initializes the class Registry."""
        self._registry = dict()

    #
    # Public functions
    #
    @_register_dependencies
    def contains(self, name: str, type: RegistryType) -> bool:
        """Indicates whether a record exists in the registry."""
        return self._contains(RegistryKey(name, type))

    @_register_dependencies
    def clear(self) -> None:
        """Clears the registry."""
        self._registry = dict()

    @_register_dependencies
    def register(self, name: str, type: RegistryType, value: Any) -> None:
        """Registers a new record."""
        registry_key = RegistryKey(name, type)
        if self._contains(registry_key):
            current_value = self.get(name=name, type=type)
            raise ValueError(
                f"Registry: `{name}` of `{type}` "
                f"is already registered as `{current_value}`."
            )
        if type in (
            RegistryType.METRICS_FUNCTION,
            RegistryType.REWARD_FUNCTION,
            RegistryType.ROLLOUT_FUNCTION,
        ) and not callable(value):
            raise ValueError(f"Registry: `{name}` of `{type}` must be callable.")

        self._registry[registry_key] = value

    @_register_dependencies
    def get(
        self,
        name: str,
        type: RegistryType,
    ) -> Optional[Callable]:
        """Gets a record by name and type."""
        registry_key = RegistryKey(name, type)
        return self._registry.get(registry_key)

    @_register_dependencies
    def get_all(self, type: RegistryType) -> dict:
        """Gets all records of a specific type."""
        return {
            key.name: value
            for key, value in self._registry.items()
            if key.registry_type == type
        }

    #
    # Convenience public function wrappers.
    #
    def get_model(self, name: str) -> Optional[Callable]:
        """Gets a record that corresponds to a registered model."""
        return self.get(name, RegistryType.MODEL)

    def get_model_config(self, name: str) -> Optional[Callable]:
        """Gets a record that corresponds to a registered config."""
        return self.get(name, RegistryType.MODEL_CONFIG)

    def get_metrics_function(self, name: str) -> Optional[Callable]:
        """Gets a record that corresponds to a registered metrics function."""
        return self.get(name, RegistryType.METRICS_FUNCTION)

    def get_reward_function(self, name: str) -> Optional[Callable]:
        """Gets a record that corresponds to a registered rewards function."""
        return self.get(name, RegistryType.REWARD_FUNCTION)

    def get_rollout_function(self, name: str) -> Optional[Callable]:
        """Gets a record that corresponds to a registered rollout function."""
        return self.get(name, RegistryType.ROLLOUT_FUNCTION)

    def get_judge_config(self, name: str) -> Optional[Callable]:
        """Gets a record that corresponds to a registered judge config."""
        return self.get(name, RegistryType.JUDGE_CONFIG)

    def get_evaluation_function(self, name: str) -> Optional[Callable]:
        """Gets a record that corresponds to a registered evaluation function."""
        return self.get(name, RegistryType.EVALUATION_FUNCTION)

    def get_sample_analyzer(self, name: str) -> Optional[Callable]:
        """Gets a record that corresponds to a registered sample analyzer."""
        return self.get(name, RegistryType.SAMPLE_ANALYZER)

    def get_dataset(
        self, name: str, subset: Optional[str] = None
    ) -> Optional[Callable]:
        """Gets a record that corresponds to a registered dataset."""
        if subset:
            # If a subset is provided, first check for subset-specific dataset.
            # If not found, try to get the dataset directly.
            dataset_cls = self.get(f"{name}/{subset}", RegistryType.DATASET)
            if dataset_cls is not None:
                return dataset_cls

        return self.get(name, RegistryType.DATASET)

    #
    # Private functions
    #
    def _contains(self, key: RegistryKey) -> bool:
        """Indicates whether a record already exists in the registry."""
        return key in self._registry

    #
    # Magic methods
    #
    def __getitem__(self, args: tuple[str, RegistryType]) -> Callable:
        """Gets a record by name and type."""
        if not isinstance(args, tuple) or len(args) != 2:
            raise ValueError(
                "Expected a tuple of length 2 with the first element being the name "
                "and the second element being the type."
            )

        name, type = args

        registry_key = RegistryKey(name, type)

        if not self._contains(registry_key):
            raise KeyError(f"Registry: `{name}` of `{type}` does not exist.")
        else:
            return self._registry[registry_key]

    def __repr__(self) -> str:
        """Defines how this class is properly printed."""
        return "\n".join(f"{key}: {value}" for key, value in self._registry.items())


REGISTRY = Registry()


def register(registry_name: str, registry_type: RegistryType) -> Callable:
    """Returns function to register decorated `obj` in the Oumi global registry.

    Args:
        registry_name: The name that the object should be registered with.
        registry_type: The type of object we are registering.

    Returns:
        Decorator function to register the target object.
    """

    def decorator_register(obj):
        """Decorator to register its target `obj`."""
        REGISTRY.register(name=registry_name, type=registry_type, value=obj)
        return obj

    return decorator_register


def register_dataset(registry_name: str, subset: Optional[str] = None) -> Callable:
    """Returns function to register decorated `obj` in the Oumi global registry.

    Args:
        registry_name: The name that the object should be registered with.
        subset: The type of object we are registering.

    Returns:
        Decorator function to register the target object.
    """

    def decorator_register(obj):
        """Decorator to register its target `obj`."""
        full_name = f"{registry_name}/{subset}" if subset else registry_name
        REGISTRY.register(name=full_name, type=RegistryType.DATASET, value=obj)
        return obj

    return decorator_register


def register_cloud_builder(registry_name: str) -> Callable:
    """Returns a function to register decorated builder in the Oumi global registry.

    Use this decorator to register cloud builder functions in the global registry.
    A cloud builder function is a function that accepts no arguments and returns an
    instance of a class that implements the `BaseCloud` interface.

    Args:
        registry_name: The name that the builder should be registered with.

    Returns:
        Decorator function to register the target builder.
    """

    def decorator_register(obj):
        """Decorator to register its target builder."""
        REGISTRY.register(name=registry_name, type=RegistryType.CLOUD, value=obj)
        return obj

    return decorator_register


def register_evaluation_function(registry_name: str) -> Callable:
    """Returns function to register an evaluation function in the Oumi global registry.

    Args:
        registry_name: The name that the evaluation function should be registered with.

    Returns:
        Decorator function to register the target evaluation function.
    """

    def check_evaluation_function_signature(evaluation_fn):
        if not callable(evaluation_fn):
            raise TypeError(
                f"Registry `{registry_name}` does not correspond to a callable object. "
                "It is required that registered evaluation functions of type "
                f"`{RegistryType.EVALUATION_FUNCTION}` must be callable."
            )

    def decorator_register(obj):
        """Decorator to register its target `obj`."""
        check_evaluation_function_signature(obj)

        REGISTRY.register(
            name=registry_name, type=RegistryType.EVALUATION_FUNCTION, value=obj
        )
        return obj

    return decorator_register


def register_sample_analyzer(registry_name: str) -> Callable:
    """Returns function to register a sample analyzer in the Oumi global registry.

    Args:
        registry_name: The name that the sample analyzer should be registered with.

    Returns:
        Decorator function to register the target sample analyzer.
    """

    def decorator_register(obj):
        """Decorator to register its target `obj`."""
        REGISTRY.register(
            name=registry_name, type=RegistryType.SAMPLE_ANALYZER, value=obj
        )
        return obj

    return decorator_register



from transformers.models.phi3.configuration_phi3 import Phi3Config

# Consider using Cambrian clone of `modeling_phi3.py`,
# which has some XLA special cases.
from transformers.models.phi3.modeling_phi3 import Phi3ForCausalLM, Phi3Model



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Collects sub-step/step/epoch timings."""

import copy
import pathlib
import sys
from pprint import pformat
from typing import Optional, Union

import transformers

import wandb  # isort: skip
from oumi.core.callbacks.base_trainer_callback import BaseTrainerCallback
from oumi.core.configs import TrainingParams
from oumi.core.distributed import get_device_rank_info, is_world_process_zero
from oumi.performance.telemetry import TelemetryTracker, TimerContext
from oumi.utils.device_utils import (
    log_nvidia_gpu_runtime_info,
)
from oumi.utils.io_utils import save_json
from oumi.utils.logging import logger

_LOGS_KWARG = "logs"


class TelemetryCallback(BaseTrainerCallback):
    """Trainer callback to collect sub-step/step/epoch timings.

    Based on `oumi.performance.telemetry.TelemetryTracker`.
    """

    def __init__(
        self,
        skip_first_steps: int = 1,
        world_process_zero_only: bool = True,
        include_timer_metrics: bool = False,
        track_gpu_temperature: bool = False,
        output_dir: Optional[pathlib.Path] = None,
    ):
        """Initializes the TelemetryCallback.

        Args:
            skip_first_steps: The number of initial steps to exclude from stats.
            world_process_zero_only: Whether to collect stats on the main process only.
            include_timer_metrics: Whether to add timer stats to reported metrics.
                The timings stats can be verbose/distracting, so `False` by default.
                The timings will be written to a file at the end of training regardless
                of the value of this flag.
            track_gpu_temperature:  Whether to record GPU temperature.
            output_dir: If specified, then telemetry stats will be written to
                the directory as JSON files.
        """
        self._telemetry = TelemetryTracker()
        self._microstep_timer: Optional[TimerContext] = None
        self._step_timer: Optional[TimerContext] = None
        self._epoch_timer: Optional[TimerContext] = None

        self._skip_first_steps: int = skip_first_steps
        self._include_timer_metrics = include_timer_metrics
        self._track_gpu_temperature = track_gpu_temperature
        self._output_dir: Optional[pathlib.Path] = output_dir
        self._permanently_disabled: bool = (
            world_process_zero_only and not is_world_process_zero()
        )
        self._world_process_zero_only = world_process_zero_only
        self._step: int = 0

        self._last_metrics_dict: Optional[dict[str, float]] = None

    def on_step_begin(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called at the beginning of a training step.

        If using gradient accumulation, one training step might take several inputs.
        """
        self._step += 1
        if self._callback_disabled():
            return

        self._complete_previous_microstep_if_needed()
        self._start_microstep()
        self._complete_previous_step_if_needed()
        self._start_step()

    def on_substep_end(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called at the end of a substep during gradient accumulation."""
        if self._callback_disabled():
            return

        self._complete_previous_microstep_if_needed()
        self._start_microstep()

    def on_step_end(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called at the end of each train step.

        Note that this will be called after all gradient accumulation substeps.
        """
        if self._callback_disabled():
            return

        self._complete_previous_microstep_if_needed()
        self._complete_previous_step_if_needed()
        if self._track_gpu_temperature:
            self._telemetry.record_gpu_temperature()

    def on_epoch_begin(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called at the beginning of an epoch."""
        if self._permanently_disabled:
            return

        self._complete_previous_epoch_if_needed()
        self._start_epoch()

    def on_epoch_end(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called at the end of an epoch."""
        if self._permanently_disabled:
            return
        self._complete_previous_epoch_if_needed()

        log_nvidia_gpu_runtime_info(log_prefix="On epoch end:")

    def on_log(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called after logging the last logs."""
        if self._callback_disabled():
            return

        device_rank_info = get_device_rank_info()
        basename = f"telemetry_rank{device_rank_info.rank:03}"

        summary = self._telemetry.get_summary()
        if (
            self._include_timer_metrics
            and "timers" in summary
            and _LOGS_KWARG in kwargs
        ):
            for name, stats in summary["timers"].items():
                for stats_key in ("mean", "median", "std_dev", "min", "max", "count"):
                    if stats_key in stats:
                        metric_name = f"{basename}_{name}_{stats_key}"
                        kwargs[_LOGS_KWARG][metric_name] = float(stats[stats_key])

        if (
            self._track_gpu_temperature
            and "gpu_temperature" in summary
            and summary["gpu_temperature"]
            and _LOGS_KWARG in kwargs
        ):
            stats = summary["gpu_temperature"]
            for stats_key in ("mean", "median", "std_dev", "min", "max", "count"):
                metric_name = f"{basename}_gpu_temperature_{stats_key}"
                kwargs[_LOGS_KWARG][metric_name] = float(stats[stats_key])

        if _LOGS_KWARG in kwargs and is_world_process_zero():
            self._last_metrics_dict = copy.deepcopy(kwargs[_LOGS_KWARG])

    def on_train_end(
        self,
        args: Union[transformers.TrainingArguments, TrainingParams],
        state: Optional[transformers.TrainerState] = None,
        control: Optional[transformers.TrainerControl] = None,
        **kwargs,
    ):
        """Event called at the end of training."""
        if self._callback_disabled() or not self._output_dir:
            return

        device_rank_info = get_device_rank_info()

        if is_world_process_zero():
            metrics_dict = self._last_metrics_dict or {}
            save_json(
                metrics_dict,
                self._output_dir
                / f"telemetry_callback_metrics_rank{device_rank_info.rank:04}.json",
            )
            if wandb.run:
                save_json(
                    {
                        "id": wandb.run.id,
                        "name": wandb.run.name,
                        "url": wandb.run.get_url(),
                    },
                    self._output_dir
                    / f"telemetry_callback_wandb_rank{device_rank_info.rank:04}.json",
                )

        if self._world_process_zero_only:
            if is_world_process_zero():
                summary = self._telemetry.get_summary()
                telemetry_file = (
                    self._output_dir
                    / f"telemetry_callback_rank{device_rank_info.rank:04}.json"
                )
                logger.info(f"Saving telemetry callback summary to {telemetry_file}...")
                save_json(summary, telemetry_file)
        else:
            # The function has to be called by all ranks.
            summaries = self._telemetry.get_summaries_from_all_ranks()
            if is_world_process_zero():
                summaries_dict = {
                    f"rank{rank:04}": summary for rank, summary in enumerate(summaries)
                }
                telemetry_file = self._output_dir / "telemetry_callback_all_ranks.json"
                logger.info(
                    "Saving telemetry callback summaries "
                    f"for all ranks to {telemetry_file}..."
                )
                save_json(summaries_dict, telemetry_file)

                gpu_temperature_info_dict = (
                    self._telemetry.compute_cross_rank_summaries(
                        summaries,
                        measurement_names={
                            "gpu_temperature": {"max", "mean", "median"},
                        },
                    )
                )
                logger.info(
                    f"GPU temperature summary:\n{pformat(gpu_temperature_info_dict)}"
                )
                save_json(
                    gpu_temperature_info_dict,
                    self._output_dir
                    / "telemetry_callback_gpu_temperature_summary.json",
                )

    def _callback_disabled(self) -> bool:
        """Check if the callback should be disabled."""
        if self._permanently_disabled:
            return True
        if self._skip_first_steps > 0 and self._step <= self._skip_first_steps:
            return True
        return False

    @staticmethod
    def _exit_timer_if_needed(timer: Optional[TimerContext]) -> Optional[TimerContext]:
        if timer is not None:
            timer.__exit__(*sys.exc_info())
        return None

    def _start_timer(self, timer_name: str) -> TimerContext:
        timer: TimerContext = self._telemetry.timer(timer_name)
        timer.__enter__()
        return timer

    def _complete_previous_microstep_if_needed(self):
        self._microstep_timer = TelemetryCallback._exit_timer_if_needed(
            self._microstep_timer
        )

    def _start_microstep(self):
        self._microstep_timer = self._start_timer("microsteps")

    def _complete_previous_step_if_needed(self):
        self._step_timer = TelemetryCallback._exit_timer_if_needed(self._step_timer)

    def _start_step(self):
        self._step_timer = self._start_timer("steps")

    def _complete_previous_epoch_if_needed(self):
        self._epoch_timer = TelemetryCallback._exit_timer_if_needed(self._epoch_timer)

    def _start_epoch(self):
        self._epoch_timer = self._start_timer("epochs")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional

from typing_extensions import override

from oumi.core.configs import GenerationParams, ModelParams, RemoteParams
from oumi.core.types.conversation import Conversation, Message, Role
from oumi.inference.remote_inference_engine import RemoteInferenceEngine


class SambanovaInferenceEngine(RemoteInferenceEngine):
    """Engine for running inference against the SambaNova API.

    This class extends RemoteInferenceEngine to provide specific functionality
    for interacting with SambaNova's language models via their API. It handles
    the conversion of Oumi's Conversation objects to SambaNova's expected input
    format, as well as parsing the API responses back into Conversation objects.
    """

    @property
    @override
    def base_url(self) -> Optional[str]:
        """Return the default base URL for the SambaNova API."""
        return "https://api.sambanova.ai/v1/chat/completions"

    @property
    @override
    def api_key_env_varname(self) -> Optional[str]:
        """Return the default environment variable name for the SambaNova API key."""
        return "SAMBANOVA_API_KEY"

    @override
    def _convert_conversation_to_api_input(
        self,
        conversation: Conversation,
        generation_params: GenerationParams,
        model_params: ModelParams,
    ) -> dict[str, Any]:
        """Converts a conversation to a SambaNova API input.

        This method transforms an Oumi Conversation object into a format
        suitable for the SambaNova API. It handles the conversion of messages
        and generation parameters according to the API specification.

        Args:
            conversation: The Oumi Conversation object to convert.
            generation_params: Parameters for text generation.
            model_params: Model parameters to use during inference.

        Returns:
            Dict[str, Any]: A dictionary containing the formatted input for the
            SambaNova API, including the model, messages, and generation parameters.
        """
        # Build request body according to SambaNova API spec
        body = {
            "model": model_params.model_name,
            "messages": self._get_list_of_message_json_dicts(
                conversation.messages, group_adjacent_same_role_turns=False
            ),
            "max_tokens": generation_params.max_new_tokens,
            "temperature": generation_params.temperature,
            "top_p": generation_params.top_p,
            "stream": False,  # We don't support streaming yet
        }

        if generation_params.stop_strings:
            body["stop"] = generation_params.stop_strings

        return body

    @override
    def _convert_api_output_to_conversation(
        self, response: dict[str, Any], original_conversation: Conversation
    ) -> Conversation:
        """Converts a SambaNova API response to a conversation.

        Args:
            response: The API response to convert.
            original_conversation: The original conversation.

        Returns:
            Conversation: The conversation including the generated response.
        """
        choices = response.get("choices", [])
        if not choices:
            raise RuntimeError("No choices found in API response")
        if len(choices) != 1:
            raise RuntimeError(
                "Sambanova API only supports one choice per response. "
                f"Got: {len(choices)}"
            )

        message = choices[0].get("message", {})
        if not message:
            raise RuntimeError("No message found in API response")

        new_message = Message(
            content=message.get("content", ""),
            role=Role.ASSISTANT,
        )

        return Conversation(
            messages=[*original_conversation.messages, new_message],
            metadata=original_conversation.metadata,
            conversation_id=original_conversation.conversation_id,
        )

    @override
    def _get_request_headers(self, remote_params: RemoteParams) -> dict[str, str]:
        """Get headers for the API request.

        Args:
            remote_params: Remote server parameters.

        Returns:
            Dict[str, str]: Headers for the API request.
        """
        headers = {
            "Content-Type": "application/json",
        }

        api_key = self._get_api_key(remote_params)
        if api_key:
            headers["Authorization"] = f"Bearer {api_key}"

        return headers

    @override
    def get_supported_params(self) -> set[str]:
        """Returns a set of supported generation parameters for this engine."""
        return {
            "max_new_tokens",
            "stop_strings",
            "temperature",
            "top_p",
        }



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field

from omegaconf import MISSING

from oumi.core.configs.base_config import BaseConfig
from oumi.core.configs.evaluation_config import EvaluationConfig


@dataclass
class AsyncEvaluationConfig(BaseConfig):
    evaluation: EvaluationConfig = field(default_factory=EvaluationConfig)
    """The evaluation configuration to use for each checkpoint.

    This field specifies the EvaluationConfig object that defines the parameters
    for evaluating each checkpoint. It includes settings for the dataset,
    model, generation, and evaluation framework to be used.
    """

    checkpoints_dir: str = MISSING
    """The directory to poll for new checkpoints."""

    polling_interval: float = MISSING
    """The time in seconds between the end of the previous evaluation and the start of
    the next polling attempt. Cannot be negative.
    """

    num_retries: int = 5
    """The number of times to retry polling before exiting the current job.

    A retry occurs when the job reads the target directory but cannot find a new
    model checkpoint to evaluate. Defaults to 5. Cannot be negative.
    """

    def __post_init__(self):
        """Verifies/populates params."""
        if self.polling_interval < 0:
            raise ValueError("`polling_interval` must be non-negative.")
        if self.num_retries < 0:
            raise ValueError("`num_retries` must be non-negative.")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from typing import Any, Optional

import torch

from oumi.core.collators.text_collator_with_padding import TextCollatorWithPadding
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.utils.torch_utils import pad_to_max_dim_and_stack


class VisionLanguageCollatorWithPadding:
    def __init__(
        self,
        tokenizer: BaseTokenizer,
        *,
        max_length: Optional[int],
        truncation: bool = False,
        label_ignore_index: Optional[int] = None,
        allow_multi_image_inputs: bool = True,
        main_image_feature: str = "pixel_values",
        debug: bool = False,
    ):
        """Custom collator for multi-modal vision-language training.

        Args:
        tokenizer: The tokenizer used for encoding the data.
        max_length: Padding length.
        truncation: Whether to truncate long inputs to `max_length`.
            If False, the long inputs are preserved as is even if they exceed
            `max_length`. Only has effect if `max_length` is specified.
        label_ignore_index:  If set, then label values of tokens that shouldn't
            contribute to the loss computation will be replaced by this special value.
        allow_multi_image_inputs: Whether to allow multi-image inputs.
        main_image_feature: The key to use for fetching the main image data
        (e.g., raw pixels, patches, etc.) from the input.
        debug: Whether to log a debug example.
        """
        self._allow_multi_image_inputs = allow_multi_image_inputs
        self._main_image_feature = main_image_feature
        self._text_collator: TextCollatorWithPadding = TextCollatorWithPadding(
            tokenizer=tokenizer,
            max_length=max_length,
            truncation=truncation,
            label_ignore_index=label_ignore_index,
            max_variable_sized_dims=(
                # if multi-image inputs are possible, then
                # allow 2 variable-sized dimensions: `seq_len`, `num_images`.
                2 if allow_multi_image_inputs else 1
            ),
            debug=debug,
        )

    def __call__(self, batch) -> dict[str, Any]:
        """Custom collator for multi-modal vision-language training.

        Args:
            batch: List of batch items.

        Returns:
            Dict[str, torch.Tensor]: Processed batch.
        """
        # Collate batch prompts
        collated_batch = self._text_collator(batch)  # type: ignore
        known_input_names: set[str] = set(collated_batch.keys()).union(
            {self._main_image_feature}
        )
        other_input_names: set[str] = set()

        images = []
        for item in batch:
            # TODO Consider relaxing this constraint: a vision/language model
            # can handle text-only inputs e.g., a follow-up to an answer,
            # or image-only inputs e.g., captioning.
            if self._main_image_feature not in item:
                raise ValueError(
                    f"Item doesn't contain '{self._main_image_feature}' key. "
                    f"Available keys: {item.keys()}"
                )
            images.append(item[self._main_image_feature])

            for key in item:
                if (
                    key
                    and (key not in known_input_names)
                    and (key not in other_input_names)
                ):
                    other_input_names.add(key)

        # Collate images.
        image_input_features = self.collate_images(images)

        # Add images to other inputs.
        collated_batch[self._main_image_feature] = image_input_features

        # For other inputs, let's verify they present in all examples and stack them.
        if len(other_input_names) > 0:
            other_inputs: dict[str, list[Any]] = collections.defaultdict(list)
            for item in batch:
                for input_name in other_input_names:
                    if input_name not in item:
                        raise ValueError(
                            f"Item doesn't contain '{input_name}' key. "
                            f"Available keys: {item.keys()}"
                        )
                    other_inputs[input_name].append(item[input_name])

            for input_name, values_list in other_inputs.items():
                collated_value = pad_to_max_dim_and_stack(
                    values_list,
                    max_variable_sized_dims=(
                        # if multi-image inputs are possible, then
                        # allow 1 variable-sized dimension (`num_images`).
                        1 if self._allow_multi_image_inputs else 0
                    ),
                )
                collated_batch[input_name] = collated_value

        return collated_batch

    def collate_images(self, images) -> torch.Tensor:
        """Collate images for multi-modal training.

        Args:
            images: List of images to collate.

        Returns:
            torch.Tensor: Batch of processed images.
        """
        if len(images) == 0:
            raise ValueError("No images found in the batch")

        return pad_to_max_dim_and_stack(
            images,
            max_variable_sized_dims=(
                # if multi-image inputs are possible, then
                # allow 1 variable-sized dimension (`num_images`).
                1 if self._allow_multi_image_inputs else 0
            ),
        )
